# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: cogment/api/agent.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from cogment.api import common_pb2 as cogment_dot_api_dot_common__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x17\x63ogment/api/agent.proto\x12\ncogmentAPI\x1a\x18\x63ogment/api/common.proto2\xa5\x01\n\x0eServiceActorSP\x12Q\n\x08RunTrial\x12\x1e.cogmentAPI.ActorRunTrialInput\x1a\x1f.cogmentAPI.ActorRunTrialOutput\"\x00(\x01\x30\x01\x12@\n\x07Version\x12\x1a.cogmentAPI.VersionRequest\x1a\x17.cogmentAPI.VersionInfo\"\x00\x62\x06proto3')



_SERVICEACTORSP = DESCRIPTOR.services_by_name['ServiceActorSP']
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _SERVICEACTORSP._serialized_start=66
  _SERVICEACTORSP._serialized_end=231
# @@protoc_insertion_point(module_scope)



# Copyright 2021 AI Redefined Inc. <dev+cogment@ai-r.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from cogment.errors import CogmentError


class Trial:
    """Internal class maintaining the information of a trial."""

    class Actor:
        """Internal class maintaining the information of an actor in a trial."""
        def __init__(self, name, actor_class):
            self.name = name
            self.actor_class = actor_class

    def __init__(self, id, actors_in_trial, cog_settings):
        self.id = id
        self.ended = False
        self.ending = False
        self.ending_ack = False
        self.cog_settings = cog_settings
        self.tick_id = -1

        self.actors = []
        for actor in actors_in_trial:
            if actor.actor_class not in self.cog_settings.actor_classes:
                raise CogmentError(f"class [{actor.actor_class}] of actor [{actor.name}] cannot be found.")
            actor_class = self.cog_settings.actor_classes[actor.actor_class]
            new_actor = self.Actor(name=actor.name, actor_class=actor_class)
            self.actors.append(new_actor)

    def __str__(self):
        result = f"Trial: id = {self.id}, tick_id = {self.tick_id}, ended = {self.ended}"
        result += f", actors = "
        for actor in self.actors:
            result += f"{{name = {actor.name}, class = {actor.actor_class.name}}},"
        return result



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc

from cogment.api import common_pb2 as cogment_dot_api_dot_common__pb2
from cogment.api import directory_pb2 as cogment_dot_api_dot_directory__pb2


class DirectorySPStub(object):
    """Used for directory (service discovery) and first line load balancing.
    """

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.Register = channel.stream_stream(
                '/cogmentAPI.DirectorySP/Register',
                request_serializer=cogment_dot_api_dot_directory__pb2.RegisterRequest.SerializeToString,
                response_deserializer=cogment_dot_api_dot_directory__pb2.RegisterReply.FromString,
                )
        self.Deregister = channel.stream_stream(
                '/cogmentAPI.DirectorySP/Deregister',
                request_serializer=cogment_dot_api_dot_directory__pb2.DeregisterRequest.SerializeToString,
                response_deserializer=cogment_dot_api_dot_directory__pb2.DeregisterReply.FromString,
                )
        self.Inquire = channel.unary_stream(
                '/cogmentAPI.DirectorySP/Inquire',
                request_serializer=cogment_dot_api_dot_directory__pb2.InquireRequest.SerializeToString,
                response_deserializer=cogment_dot_api_dot_directory__pb2.InquireReply.FromString,
                )
        self.Version = channel.unary_unary(
                '/cogmentAPI.DirectorySP/Version',
                request_serializer=cogment_dot_api_dot_common__pb2.VersionRequest.SerializeToString,
                response_deserializer=cogment_dot_api_dot_common__pb2.VersionInfo.FromString,
                )


class DirectorySPServicer(object):
    """Used for directory (service discovery) and first line load balancing.
    """

    def Register(self, request_iterator, context):
        """Register as a service.
        - There will be a one-to-one match between requests and replies
        - Each request can succeed or fail individually
        Expected metadata:
        - authentication-token (may be optional, at the discretion of the implementation).
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Deregister(self, request_iterator, context):
        """Deregister a service.
        - There will be a one-to-one match between requests and replies
        - Each request can succeed or fail individually
        Expected metadata:
        - authentication-token (required if it was used to register).
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Inquire(self, request, context):
        """Inquire about registered services.
        Expected metadata:
        - authentication-token (may be optional, at the discretion of the implemetation).
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Version(self, request, context):
        """Expected metadata: None
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_DirectorySPServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'Register': grpc.stream_stream_rpc_method_handler(
                    servicer.Register,
                    request_deserializer=cogment_dot_api_dot_directory__pb2.RegisterRequest.FromString,
                    response_serializer=cogment_dot_api_dot_directory__pb2.RegisterReply.SerializeToString,
            ),
            'Deregister': grpc.stream_stream_rpc_method_handler(
                    servicer.Deregister,
                    request_deserializer=cogment_dot_api_dot_directory__pb2.DeregisterRequest.FromString,
                    response_serializer=cogment_dot_api_dot_directory__pb2.DeregisterReply.SerializeToString,
            ),
            'Inquire': grpc.unary_stream_rpc_method_handler(
                    servicer.Inquire,
                    request_deserializer=cogment_dot_api_dot_directory__pb2.InquireRequest.FromString,
                    response_serializer=cogment_dot_api_dot_directory__pb2.InquireReply.SerializeToString,
            ),
            'Version': grpc.unary_unary_rpc_method_handler(
                    servicer.Version,
                    request_deserializer=cogment_dot_api_dot_common__pb2.VersionRequest.FromString,
                    response_serializer=cogment_dot_api_dot_common__pb2.VersionInfo.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'cogmentAPI.DirectorySP', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))


 # This class is part of an EXPERIMENTAL API.
class DirectorySP(object):
    """Used for directory (service discovery) and first line load balancing.
    """

    @staticmethod
    def Register(request_iterator,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.stream_stream(request_iterator, target, '/cogmentAPI.DirectorySP/Register',
            cogment_dot_api_dot_directory__pb2.RegisterRequest.SerializeToString,
            cogment_dot_api_dot_directory__pb2.RegisterReply.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def Deregister(request_iterator,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.stream_stream(request_iterator, target, '/cogmentAPI.DirectorySP/Deregister',
            cogment_dot_api_dot_directory__pb2.DeregisterRequest.SerializeToString,
            cogment_dot_api_dot_directory__pb2.DeregisterReply.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def Inquire(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_stream(request, target, '/cogmentAPI.DirectorySP/Inquire',
            cogment_dot_api_dot_directory__pb2.InquireRequest.SerializeToString,
            cogment_dot_api_dot_directory__pb2.InquireReply.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def Version(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/cogmentAPI.DirectorySP/Version',
            cogment_dot_api_dot_common__pb2.VersionRequest.SerializeToString,
            cogment_dot_api_dot_common__pb2.VersionInfo.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc

from cogment.api import common_pb2 as cogment_dot_api_dot_common__pb2


class ServiceActorSPStub(object):
    """Missing associated documentation comment in .proto file."""

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.RunTrial = channel.stream_stream(
                '/cogmentAPI.ServiceActorSP/RunTrial',
                request_serializer=cogment_dot_api_dot_common__pb2.ActorRunTrialInput.SerializeToString,
                response_deserializer=cogment_dot_api_dot_common__pb2.ActorRunTrialOutput.FromString,
                )
        self.Version = channel.unary_unary(
                '/cogmentAPI.ServiceActorSP/Version',
                request_serializer=cogment_dot_api_dot_common__pb2.VersionRequest.SerializeToString,
                response_deserializer=cogment_dot_api_dot_common__pb2.VersionInfo.FromString,
                )


class ServiceActorSPServicer(object):
    """Missing associated documentation comment in .proto file."""

    def RunTrial(self, request_iterator, context):
        """Expected metadata:
        - trial-id: The id of the trial
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Version(self, request, context):
        """Expected metadata: None
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_ServiceActorSPServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'RunTrial': grpc.stream_stream_rpc_method_handler(
                    servicer.RunTrial,
                    request_deserializer=cogment_dot_api_dot_common__pb2.ActorRunTrialInput.FromString,
                    response_serializer=cogment_dot_api_dot_common__pb2.ActorRunTrialOutput.SerializeToString,
            ),
            'Version': grpc.unary_unary_rpc_method_handler(
                    servicer.Version,
                    request_deserializer=cogment_dot_api_dot_common__pb2.VersionRequest.FromString,
                    response_serializer=cogment_dot_api_dot_common__pb2.VersionInfo.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'cogmentAPI.ServiceActorSP', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))


 # This class is part of an EXPERIMENTAL API.
class ServiceActorSP(object):
    """Missing associated documentation comment in .proto file."""

    @staticmethod
    def RunTrial(request_iterator,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.stream_stream(request_iterator, target, '/cogmentAPI.ServiceActorSP/RunTrial',
            cogment_dot_api_dot_common__pb2.ActorRunTrialInput.SerializeToString,
            cogment_dot_api_dot_common__pb2.ActorRunTrialOutput.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def Version(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/cogmentAPI.ServiceActorSP/Version',
            cogment_dot_api_dot_common__pb2.VersionRequest.SerializeToString,
            cogment_dot_api_dot_common__pb2.VersionInfo.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: cogment/api/directory.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from cogment.api import common_pb2 as cogment_dot_api_dot_common__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x1b\x63ogment/api/directory.proto\x12\ncogmentAPI\x1a\x18\x63ogment/api/common.proto\"m\n\x0fRegisterRequest\x12-\n\x08\x65ndpoint\x18\x01 \x01(\x0b\x32\x1b.cogmentAPI.ServiceEndpoint\x12+\n\x07\x64\x65tails\x18\x02 \x01(\x0b\x32\x1a.cogmentAPI.ServiceDetails\"\xa3\x01\n\rRegisterReply\x12\x30\n\x06status\x18\x01 \x01(\x0e\x32 .cogmentAPI.RegisterReply.Status\x12\x11\n\terror_msg\x18\x02 \x01(\t\x12\x12\n\nservice_id\x18\x03 \x01(\x04\x12\x0e\n\x06secret\x18\x04 \x01(\t\")\n\x06Status\x12\x0b\n\x07UNKNOWN\x10\x00\x12\x06\n\x02OK\x10\x01\x12\n\n\x06\x46\x41ILED\x10\x02\"7\n\x11\x44\x65registerRequest\x12\x12\n\nservice_id\x18\x01 \x01(\x04\x12\x0e\n\x06secret\x18\x02 \x01(\t\"\x83\x01\n\x0f\x44\x65registerReply\x12\x32\n\x06status\x18\x01 \x01(\x0e\x32\".cogmentAPI.DeregisterReply.Status\x12\x11\n\terror_msg\x18\x02 \x01(\t\")\n\x06Status\x12\x0b\n\x07UNKNOWN\x10\x00\x12\x06\n\x02OK\x10\x01\x12\n\n\x06\x46\x41ILED\x10\x02\"`\n\x0eInquireRequest\x12\x14\n\nservice_id\x18\x01 \x01(\x04H\x00\x12-\n\x07\x64\x65tails\x18\x02 \x01(\x0b\x32\x1a.cogmentAPI.ServiceDetailsH\x00\x42\t\n\x07inquiry\"9\n\x0cInquireReply\x12)\n\x04\x64\x61ta\x18\x01 \x01(\x0b\x32\x1b.cogmentAPI.FullServiceData\"\xa3\x01\n\x0fServiceEndpoint\x12\x36\n\x08protocol\x18\x01 \x01(\x0e\x32$.cogmentAPI.ServiceEndpoint.Protocol\x12\x0c\n\x04host\x18\x02 \x01(\t\x12\x0c\n\x04port\x18\x03 \x01(\r\"<\n\x08Protocol\x12\x0b\n\x07UNKNOWN\x10\x00\x12\x08\n\x04GRPC\x10\x01\x12\x0c\n\x08GRPC_SSL\x10\x02\x12\x0b\n\x07\x43OGMENT\x10\x03\"\xaa\x01\n\x0eServiceDetails\x12%\n\x04type\x18\x01 \x01(\x0e\x32\x17.cogmentAPI.ServiceType\x12>\n\nproperties\x18\x02 \x03(\x0b\x32*.cogmentAPI.ServiceDetails.PropertiesEntry\x1a\x31\n\x0fPropertiesEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\"\x81\x01\n\x0f\x46ullServiceData\x12-\n\x08\x65ndpoint\x18\x01 \x01(\x0b\x32\x1b.cogmentAPI.ServiceEndpoint\x12\x12\n\nservice_id\x18\x02 \x01(\x04\x12+\n\x07\x64\x65tails\x18\x03 \x01(\x0b\x32\x1a.cogmentAPI.ServiceDetails*\x82\x02\n\x0bServiceType\x12\x13\n\x0fUNKNOWN_SERVICE\x10\x00\x12\x1c\n\x18TRIAL_LIFE_CYCLE_SERVICE\x10\x01\x12#\n\x1f\x43LIENT_ACTOR_CONNECTION_SERVICE\x10\x02\x12\x11\n\rACTOR_SERVICE\x10\x03\x12\x17\n\x13\x45NVIRONMENT_SERVICE\x10\x04\x12\x14\n\x10PRE_HOOK_SERVICE\x10\x05\x12\x13\n\x0f\x44\x41TALOG_SERVICE\x10\x06\x12\x15\n\x11\x44\x41TASTORE_SERVICE\x10\x07\x12\x1a\n\x16MODEL_REGISTRY_SERVICE\x10\x08\x12\x11\n\rOTHER_SERVICE\x10\x64\x32\xae\x02\n\x0b\x44irectorySP\x12H\n\x08Register\x12\x1b.cogmentAPI.RegisterRequest\x1a\x19.cogmentAPI.RegisterReply\"\x00(\x01\x30\x01\x12N\n\nDeregister\x12\x1d.cogmentAPI.DeregisterRequest\x1a\x1b.cogmentAPI.DeregisterReply\"\x00(\x01\x30\x01\x12\x43\n\x07Inquire\x12\x1a.cogmentAPI.InquireRequest\x1a\x18.cogmentAPI.InquireReply\"\x00\x30\x01\x12@\n\x07Version\x12\x1a.cogmentAPI.VersionRequest\x1a\x17.cogmentAPI.VersionInfo\"\x00\x62\x06proto3')

_SERVICETYPE = DESCRIPTOR.enum_types_by_name['ServiceType']
ServiceType = enum_type_wrapper.EnumTypeWrapper(_SERVICETYPE)
UNKNOWN_SERVICE = 0
TRIAL_LIFE_CYCLE_SERVICE = 1
CLIENT_ACTOR_CONNECTION_SERVICE = 2
ACTOR_SERVICE = 3
ENVIRONMENT_SERVICE = 4
PRE_HOOK_SERVICE = 5
DATALOG_SERVICE = 6
DATASTORE_SERVICE = 7
MODEL_REGISTRY_SERVICE = 8
OTHER_SERVICE = 100


_REGISTERREQUEST = DESCRIPTOR.message_types_by_name['RegisterRequest']
_REGISTERREPLY = DESCRIPTOR.message_types_by_name['RegisterReply']
_DEREGISTERREQUEST = DESCRIPTOR.message_types_by_name['DeregisterRequest']
_DEREGISTERREPLY = DESCRIPTOR.message_types_by_name['DeregisterReply']
_INQUIREREQUEST = DESCRIPTOR.message_types_by_name['InquireRequest']
_INQUIREREPLY = DESCRIPTOR.message_types_by_name['InquireReply']
_SERVICEENDPOINT = DESCRIPTOR.message_types_by_name['ServiceEndpoint']
_SERVICEDETAILS = DESCRIPTOR.message_types_by_name['ServiceDetails']
_SERVICEDETAILS_PROPERTIESENTRY = _SERVICEDETAILS.nested_types_by_name['PropertiesEntry']
_FULLSERVICEDATA = DESCRIPTOR.message_types_by_name['FullServiceData']
_REGISTERREPLY_STATUS = _REGISTERREPLY.enum_types_by_name['Status']
_DEREGISTERREPLY_STATUS = _DEREGISTERREPLY.enum_types_by_name['Status']
_SERVICEENDPOINT_PROTOCOL = _SERVICEENDPOINT.enum_types_by_name['Protocol']
RegisterRequest = _reflection.GeneratedProtocolMessageType('RegisterRequest', (_message.Message,), {
  'DESCRIPTOR' : _REGISTERREQUEST,
  '__module__' : 'cogment.api.directory_pb2'
  # @@protoc_insertion_point(class_scope:cogmentAPI.RegisterRequest)
  })
_sym_db.RegisterMessage(RegisterRequest)

RegisterReply = _reflection.GeneratedProtocolMessageType('RegisterReply', (_message.Message,), {
  'DESCRIPTOR' : _REGISTERREPLY,
  '__module__' : 'cogment.api.directory_pb2'
  # @@protoc_insertion_point(class_scope:cogmentAPI.RegisterReply)
  })
_sym_db.RegisterMessage(RegisterReply)

DeregisterRequest = _reflection.GeneratedProtocolMessageType('DeregisterRequest', (_message.Message,), {
  'DESCRIPTOR' : _DEREGISTERREQUEST,
  '__module__' : 'cogment.api.directory_pb2'
  # @@protoc_insertion_point(class_scope:cogmentAPI.DeregisterRequest)
  })
_sym_db.RegisterMessage(DeregisterRequest)

DeregisterReply = _reflection.GeneratedProtocolMessageType('DeregisterReply', (_message.Message,), {
  'DESCRIPTOR' : _DEREGISTERREPLY,
  '__module__' : 'cogment.api.directory_pb2'
  # @@protoc_insertion_point(class_scope:cogmentAPI.DeregisterReply)
  })
_sym_db.RegisterMessage(DeregisterReply)

InquireRequest = _reflection.GeneratedProtocolMessageType('InquireRequest', (_message.Message,), {
  'DESCRIPTOR' : _INQUIREREQUEST,
  '__module__' : 'cogment.api.directory_pb2'
  # @@protoc_insertion_point(class_scope:cogmentAPI.InquireRequest)
  })
_sym_db.RegisterMessage(InquireRequest)

InquireReply = _reflection.GeneratedProtocolMessageType('InquireReply', (_message.Message,), {
  'DESCRIPTOR' : _INQUIREREPLY,
  '__module__' : 'cogment.api.directory_pb2'
  # @@protoc_insertion_point(class_scope:cogmentAPI.InquireReply)
  })
_sym_db.RegisterMessage(InquireReply)

ServiceEndpoint = _reflection.GeneratedProtocolMessageType('ServiceEndpoint', (_message.Message,), {
  'DESCRIPTOR' : _SERVICEENDPOINT,
  '__module__' : 'cogment.api.directory_pb2'
  # @@protoc_insertion_point(class_scope:cogmentAPI.ServiceEndpoint)
  })
_sym_db.RegisterMessage(ServiceEndpoint)

ServiceDetails = _reflection.GeneratedProtocolMessageType('ServiceDetails', (_message.Message,), {

  'PropertiesEntry' : _reflection.GeneratedProtocolMessageType('PropertiesEntry', (_message.Message,), {
    'DESCRIPTOR' : _SERVICEDETAILS_PROPERTIESENTRY,
    '__module__' : 'cogment.api.directory_pb2'
    # @@protoc_insertion_point(class_scope:cogmentAPI.ServiceDetails.PropertiesEntry)
    })
  ,
  'DESCRIPTOR' : _SERVICEDETAILS,
  '__module__' : 'cogment.api.directory_pb2'
  # @@protoc_insertion_point(class_scope:cogmentAPI.ServiceDetails)
  })
_sym_db.RegisterMessage(ServiceDetails)
_sym_db.RegisterMessage(ServiceDetails.PropertiesEntry)

FullServiceData = _reflection.GeneratedProtocolMessageType('FullServiceData', (_message.Message,), {
  'DESCRIPTOR' : _FULLSERVICEDATA,
  '__module__' : 'cogment.api.directory_pb2'
  # @@protoc_insertion_point(class_scope:cogmentAPI.FullServiceData)
  })
_sym_db.RegisterMessage(FullServiceData)

_DIRECTORYSP = DESCRIPTOR.services_by_name['DirectorySP']
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _SERVICEDETAILS_PROPERTIESENTRY._options = None
  _SERVICEDETAILS_PROPERTIESENTRY._serialized_options = b'8\001'
  _SERVICETYPE._serialized_start=1166
  _SERVICETYPE._serialized_end=1424
  _REGISTERREQUEST._serialized_start=69
  _REGISTERREQUEST._serialized_end=178
  _REGISTERREPLY._serialized_start=181
  _REGISTERREPLY._serialized_end=344
  _REGISTERREPLY_STATUS._serialized_start=303
  _REGISTERREPLY_STATUS._serialized_end=344
  _DEREGISTERREQUEST._serialized_start=346
  _DEREGISTERREQUEST._serialized_end=401
  _DEREGISTERREPLY._serialized_start=404
  _DEREGISTERREPLY._serialized_end=535
  _DEREGISTERREPLY_STATUS._serialized_start=303
  _DEREGISTERREPLY_STATUS._serialized_end=344
  _INQUIREREQUEST._serialized_start=537
  _INQUIREREQUEST._serialized_end=633
  _INQUIREREPLY._serialized_start=635
  _INQUIREREPLY._serialized_end=692
  _SERVICEENDPOINT._serialized_start=695
  _SERVICEENDPOINT._serialized_end=858
  _SERVICEENDPOINT_PROTOCOL._serialized_start=798
  _SERVICEENDPOINT_PROTOCOL._serialized_end=858
  _SERVICEDETAILS._serialized_start=861
  _SERVICEDETAILS._serialized_end=1031
  _SERVICEDETAILS_PROPERTIESENTRY._serialized_start=982
  _SERVICEDETAILS_PROPERTIESENTRY._serialized_end=1031
  _FULLSERVICEDATA._serialized_start=1034
  _FULLSERVICEDATA._serialized_end=1163
  _DIRECTORYSP._serialized_start=1427
  _DIRECTORYSP._serialized_end=1729
# @@protoc_insertion_point(module_scope)



# Copyright 2021 AI Redefined Inc. <dev+cogment@ai-r.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import cogment.api.common_pb2 as common_api

from cogment.errors import CogmentError
from cogment.utils import logger

from abc import ABC
from enum import Enum
import asyncio


# This class is not necessary, but simplifies the code as it is.
# I.e. the actual init_output proto message could be sent instead.
class _InitAck:
    """Internal class to signal acknowledgement of the start."""
    pass


class _Ending:
    """Internal class to signal the ending."""
    pass


class _EndingAck:
    """Internal class to signal acknowledgement of the ending."""
    pass


class ActorInfo:
    """Class representing the information of an actor."""

    def __init__(self, name, class_name):
        self.actor_name = name
        self.actor_class_name = class_name

    def __str__(self):
        result = f"ActorInfo: actor_name = {self.actor_name}, actor_class_name = {self.actor_class_name}"
        return result


class RecvObservation:
    """Class representing a received observation in a trial."""

    def __init__(self, obs, obs_space):
        self.tick_id = obs.tick_id
        self.timestamp = obs.timestamp
        self.observation = obs_space

    def __str__(self):
        result = f"RecvObservation: tick_id = {self.tick_id}, timestamp = {self.timestamp}"
        result += f", observation = {self.observation}"
        return result

    @property
    def snapshot(self):
        logger.deprecated(f"Deprecated use of 'snapshot' in RecvObservation. Use 'observation' instead.")
        return self.observation

    @snapshot.setter
    def snapshot(self, val):
        logger.deprecated(f"Deprecated use of 'snapshot' in RecvObservation. Use 'observation' instead.")
        self.observation = val


class ActorStatus(Enum):
    """Enum class for the status of actors in a trial."""

    UNKNOWN = 0
    ACTIVE = 1
    UNAVAILABLE = 2
    DEFAULT = 3


class RecvAction:
    """Class representing a received action in a trial."""

    def __init__(self, actor_index, tick_id, status, timestamp, action):
        self.tick_id = tick_id
        self.actor_index = actor_index
        self.status = status
        self.timestamp = timestamp
        self.action = action

    def __str__(self):
        result = f"RecvAction: tick_id = {self.tick_id}, actor_index = {self.actor_index}"
        result += f", status = {self.status}, timestamp = {self.timestamp}, action = {self.action}"
        return result


class RecvRewardSource:
    """Class representing a received reward source in a trial."""

    def __init__(self, src):
        self.value = src.value
        self.confidence = src.confidence
        self.sender_name = src.sender_name
        self.user_data = src.user_data

    def __str__(self):
        result = f"RecvRewardSource: value = {self.value}, confidence = {self.confidence}"
        result += f", sender_name = {self.sender_name}, user_data = {self.user_data}"
        return result


class RecvReward:
    """Class representing a received reward in a trial."""

    def __init__(self, reward):
        self.receiver_name = reward.receiver_name
        self.tick_id = reward.tick_id
        self.value = reward.value
        self._sources = reward.sources

    def get_nb_sources(self):
        return len(self._sources)

    def all_sources(self):
        if not self._sources:
            raise CogmentError("Unexpected reward with no source")
        for src in self._sources:
            yield RecvRewardSource(src)

    def __str__(self):
        result = f"RecvReward: tick_id = {self.tick_id}, receiver_name = {self.receiver_name}"
        result += f", value = {self.value}, sources = {self._sources}"
        return result


class RecvMessage:
    """Class representing a received message in a trial."""

    def __init__(self, message):
        self.receiver_name = message.receiver_name
        self.tick_id = message.tick_id
        self.sender_name = message.sender_name
        self.payload = message.payload

    def __str__(self):
        result = f"RecvMessage: tick_id = {self.tick_id}, receiver_name = {self.receiver_name}"
        result += f", sender_name = {self.sender_name}, payload = {self.payload}"
        return result


class EventType(Enum):
    """Enum class for the types of received events in a trial."""

    NONE = 0
    ACTIVE = 1
    ENDING = 2
    FINAL = 3


class RecvEvent:
    """Class representing a received event in a trial."""

    def __init__(self, etype):
        self.type = etype
        self.observation = None
        self.actions = []
        self.rewards = []
        self.messages = []

    def __str__(self):
        result = f"RecvEvent: type = {self.type}"
        if self.observation:
            result += f", {{{self.observation}}}"
        for act in self.actions:
            result += f", {{{act}}}"
        for rew in self.rewards:
            result += f", {{{rew}}}"
        for msg in self.messages:
            result += f", {{{msg}}}"
        return result


class Session(ABC):
    """Base class representing the session of an actor or environment for a trial."""

    def __init__(self, trial, name, impl, impl_name, config):
        self._trial = trial
        self.name = name
        self.impl_name = impl_name
        self._impl = impl
        self.config = config
        self._incoming_event_queue = asyncio.Queue()
        self._outgoing_data_queue = asyncio.Queue()
        self._started = False
        self._last_tick_delivered = -1
        self._last_event_delivered = False
        self._user_task = None  # Task used to call user implementation
        self._auto_ack = True

        # Pre-compute since it will be used regularly
        self._active_actors = [ActorInfo(actor.name, actor.actor_class.name) for actor in trial.actors]

    def __str__(self):
        result = f"Session: name = {self.name}, impl_name = {self.impl_name}, config = {self.config}"
        result += f", trial = {self._trial}"
        return result

    def _start(self, auto_done_sending):
        if self._started:
            raise CogmentError(f"Cannot start [{self.name}] more than once.")
        if self._trial.ended:
            raise CogmentError(f"Cannot start [{self.name}] because the trial has ended.")

        self._auto_ack = auto_done_sending
        self._started = True
        self._post_outgoing_data(_InitAck())

    def _exit_queues(self):
        self._incoming_event_queue.put_nowait(None)
        self._outgoing_data_queue.put_nowait(None)

    async def _run(self):
        try:
            await self._impl(self)
            return True

        except asyncio.CancelledError as exc:
            logger.debug(f"[{self.name}] implementation coroutine cancelled: [{exc}]")
            return False

        except Exception:
            logger.exception(f"An exception occured in user implementation of [{self.name}]")
            raise

    def _start_user_task(self):
        self._user_task = asyncio.create_task(self._run())
        return self._user_task

    def _post_incoming_event(self, event_tuple):
        """event_tuple is a tuple (tick_id, event)"""
        if not self._started:
            logger.warning(f"[{self.name}] received an event before session was started.")
            return
        if self._trial.ended:
            logger.debug(f"Event received after trial is over: [{event_tuple}]")
            return

        self._incoming_event_queue.put_nowait(event_tuple)

    def _post_outgoing_data(self, data):
        if not self._started:
            logger.warning(f"Trial [{self._trial.id}] - Session for [{self.name}]: "
                           f"Cannot send until session is started.")
            return
        if self._trial.ending_ack:
            logger.warning(f"Trial [{self._trial.id}] - Session for [{self.name}]: "
                           f"Cannot send after acknowledging ending")
            return

        if data is None:
            raise CogmentError(f"Trial [{self._trial.id}] - Session for [{self.name}]: Data posted is `None`")

        if type(data) == _Ending:
            self._trial.ending = True
        elif type(data) == _EndingAck:
            self._trial.ending_ack = True

        self._outgoing_data_queue.put_nowait(data)

    async def _retrieve_outgoing_data(self):
        try:
            while True:
                data = await self._outgoing_data_queue.get()
                if data is None:
                    logger.debug(f"Trial [{self._trial.id}] - Session for [{self.name}]: "
                                 f"Forcefull data loop exit")
                    break
                yield data
                self._outgoing_data_queue.task_done()

        except RuntimeError as exc:
            # Unfortunatelty asyncio returns a standard RuntimeError in this case
            if exc.args[0] != "Event loop is closed":
                raise
            else:
                logger.debug(f"Trial [{self._trial.id}] - Session for [{self.name}]: "
                             f"Normal exception on retrieving data at close: [{exc}]")

        except asyncio.CancelledError as exc:
            logger.debug(f"Trial [{self._trial.id}] - Session for [{self.name}]: "
                         f"data retrieval coroutine cancelled: [{exc}]")

        logger.debug(f"Exiting [{self.name}] _retrieve_outgoing_data loop generator")

    def get_trial_id(self):
        return self._trial.id

    def get_tick_id(self):
        return self._trial.tick_id

    def is_trial_over(self):
        return self._trial.ended

    def sending_done(self):
        if self._auto_ack:
            raise CogmentError("Cannot manually end sending as it is set to automatic")
        elif not self._trial.ending:
            raise CogmentError("Cannot stop sending before trial is ready to end")
        elif self._trial.ended:
            logger.warning(f"Trial [{self._trial.id}] - Session [{self.name}] "
                           f"end sending ignored because the trial has already ended.")
        elif self._trial.ending_ack:
            logger.debug(f"Trial [{self._trial.id}] - Session [{self.name}] cannot end sending more than once")
        else:
            self._post_outgoing_data(_EndingAck())

    async def all_events(self):
        if not self._started:
            logger.warning(f"Cannot retrieve events until [{self.name}] is started.")
            return
        if self._trial.ended:
            logger.info(f"No more events for [{self.name}] because the trial has ended.")
            return

        logger.debug(f"Trial [{self._trial.id}] - Session [{self.name}] starting event loop")

        loop_active = not self._last_event_delivered
        while loop_active:
            try:
                event_tuple = await self._incoming_event_queue.get()
                if event_tuple is None:
                    logger.debug(f"Trial [{self._trial.id}] - Session [{self.name}]: "
                                 f"Forcefull event loop exit")
                    self._last_event_delivered = True
                    break

            except asyncio.CancelledError as exc:
                logger.debug(f"[{self.name}] coroutine cancelled while waiting for an event: [{exc}]")
                break

            tick_id, event = event_tuple
            if tick_id >= 0:
                self._last_tick_delivered = tick_id
            self._last_event_delivered = (event.type == EventType.FINAL)

            keep_looping = yield event
            self._incoming_event_queue.task_done()

            loop_active = (keep_looping is None or bool(keep_looping)) and not self._last_event_delivered
            if not loop_active:
                if self._last_event_delivered:
                    logger.debug(f"Last event delivered, exiting [{self.name}] event loop")
                else:
                    logger.debug(f"End of event loop for [{self.name}] requested by user")

        logger.debug(f"Exiting [{self.name}] event loop generator")

    async def event_loop(self):
        logger.deprecated("`event_loop` is deprecated. Use `all_events` instead.")

        async for event in self.all_events():
            yield event

    def add_reward(self, value, confidence, to, tick_id=-1, user_data=None):
        if not self._started:
            logger.warning(f"Trial [{self._trial.id}] - Session for [{self.name}]: "
                           f"Cannot send reward until session is started.")
            return
        if self._trial.ending_ack:
            logger.warning(f"Trial [{self._trial.id}] - Session for [{self.name}]: "
                           f"Cannot send reward after acknowledging ending.")
            return

        for dest in to:
            reward = common_api.Reward(receiver_name=dest, tick_id=tick_id, value=value)
            reward_source = common_api.RewardSource(sender_name=self.name, value=value, confidence=confidence)
            if user_data is not None:
                reward_source.user_data.Pack(user_data)
            reward.sources.append(reward_source)

            self._post_outgoing_data(reward)

    def _send_message(self, payload, to):
        if not self._started:
            logger.warning(f"Trial [{self._trial.id}] - Session for [{self.name}]: "
                           f"Cannot send message until session is started.")
            return
        if self._trial.ending_ack:
            logger.warning(f"Trial [{self._trial.id}] - Session for [{self.name}]: "
                           f"Cannot send message after acknowledging ending.")
            return

        for dest in to:
            message = common_api.Message(tick_id=-1, receiver_name=dest)
            if payload is not None:
                message.payload.Pack(payload)

            self._post_outgoing_data(message)



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc

from cogment.api import common_pb2 as cogment_dot_api_dot_common__pb2
from cogment.api import orchestrator_pb2 as cogment_dot_api_dot_orchestrator__pb2


class TrialLifecycleSPStub(object):
    """This service is used to manage Trial lifecycle.
    """

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.StartTrial = channel.unary_unary(
                '/cogmentAPI.TrialLifecycleSP/StartTrial',
                request_serializer=cogment_dot_api_dot_orchestrator__pb2.TrialStartRequest.SerializeToString,
                response_deserializer=cogment_dot_api_dot_orchestrator__pb2.TrialStartReply.FromString,
                )
        self.TerminateTrial = channel.unary_unary(
                '/cogmentAPI.TrialLifecycleSP/TerminateTrial',
                request_serializer=cogment_dot_api_dot_orchestrator__pb2.TerminateTrialRequest.SerializeToString,
                response_deserializer=cogment_dot_api_dot_orchestrator__pb2.TerminateTrialReply.FromString,
                )
        self.GetTrialInfo = channel.unary_unary(
                '/cogmentAPI.TrialLifecycleSP/GetTrialInfo',
                request_serializer=cogment_dot_api_dot_orchestrator__pb2.TrialInfoRequest.SerializeToString,
                response_deserializer=cogment_dot_api_dot_orchestrator__pb2.TrialInfoReply.FromString,
                )
        self.WatchTrials = channel.unary_stream(
                '/cogmentAPI.TrialLifecycleSP/WatchTrials',
                request_serializer=cogment_dot_api_dot_orchestrator__pb2.TrialListRequest.SerializeToString,
                response_deserializer=cogment_dot_api_dot_orchestrator__pb2.TrialListEntry.FromString,
                )
        self.Version = channel.unary_unary(
                '/cogmentAPI.TrialLifecycleSP/Version',
                request_serializer=cogment_dot_api_dot_common__pb2.VersionRequest.SerializeToString,
                response_deserializer=cogment_dot_api_dot_common__pb2.VersionInfo.FromString,
                )


class TrialLifecycleSPServicer(object):
    """This service is used to manage Trial lifecycle.
    """

    def StartTrial(self, request, context):
        """Begin a new trial.
        Expected headers: None
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def TerminateTrial(self, request, context):
        """Terminate existing trial(s).
        Expected headers: 
        - trial-id <one or more>
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetTrialInfo(self, request, context):
        """Get extra information about a specific trial.
        Expected headers: 
        - trial-id <zero or more>
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def WatchTrials(self, request, context):
        """Get information about the trials currently running in the orchestrator.
        The stream will start with all trials that match the filter.
        After that, it will contain a new message whenever a trial enters a filtered state.
        Expected headers: None
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Version(self, request, context):
        """Expected headers: None
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_TrialLifecycleSPServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'StartTrial': grpc.unary_unary_rpc_method_handler(
                    servicer.StartTrial,
                    request_deserializer=cogment_dot_api_dot_orchestrator__pb2.TrialStartRequest.FromString,
                    response_serializer=cogment_dot_api_dot_orchestrator__pb2.TrialStartReply.SerializeToString,
            ),
            'TerminateTrial': grpc.unary_unary_rpc_method_handler(
                    servicer.TerminateTrial,
                    request_deserializer=cogment_dot_api_dot_orchestrator__pb2.TerminateTrialRequest.FromString,
                    response_serializer=cogment_dot_api_dot_orchestrator__pb2.TerminateTrialReply.SerializeToString,
            ),
            'GetTrialInfo': grpc.unary_unary_rpc_method_handler(
                    servicer.GetTrialInfo,
                    request_deserializer=cogment_dot_api_dot_orchestrator__pb2.TrialInfoRequest.FromString,
                    response_serializer=cogment_dot_api_dot_orchestrator__pb2.TrialInfoReply.SerializeToString,
            ),
            'WatchTrials': grpc.unary_stream_rpc_method_handler(
                    servicer.WatchTrials,
                    request_deserializer=cogment_dot_api_dot_orchestrator__pb2.TrialListRequest.FromString,
                    response_serializer=cogment_dot_api_dot_orchestrator__pb2.TrialListEntry.SerializeToString,
            ),
            'Version': grpc.unary_unary_rpc_method_handler(
                    servicer.Version,
                    request_deserializer=cogment_dot_api_dot_common__pb2.VersionRequest.FromString,
                    response_serializer=cogment_dot_api_dot_common__pb2.VersionInfo.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'cogmentAPI.TrialLifecycleSP', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))


 # This class is part of an EXPERIMENTAL API.
class TrialLifecycleSP(object):
    """This service is used to manage Trial lifecycle.
    """

    @staticmethod
    def StartTrial(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/cogmentAPI.TrialLifecycleSP/StartTrial',
            cogment_dot_api_dot_orchestrator__pb2.TrialStartRequest.SerializeToString,
            cogment_dot_api_dot_orchestrator__pb2.TrialStartReply.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def TerminateTrial(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/cogmentAPI.TrialLifecycleSP/TerminateTrial',
            cogment_dot_api_dot_orchestrator__pb2.TerminateTrialRequest.SerializeToString,
            cogment_dot_api_dot_orchestrator__pb2.TerminateTrialReply.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def GetTrialInfo(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/cogmentAPI.TrialLifecycleSP/GetTrialInfo',
            cogment_dot_api_dot_orchestrator__pb2.TrialInfoRequest.SerializeToString,
            cogment_dot_api_dot_orchestrator__pb2.TrialInfoReply.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def WatchTrials(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_stream(request, target, '/cogmentAPI.TrialLifecycleSP/WatchTrials',
            cogment_dot_api_dot_orchestrator__pb2.TrialListRequest.SerializeToString,
            cogment_dot_api_dot_orchestrator__pb2.TrialListEntry.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def Version(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/cogmentAPI.TrialLifecycleSP/Version',
            cogment_dot_api_dot_common__pb2.VersionRequest.SerializeToString,
            cogment_dot_api_dot_common__pb2.VersionInfo.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)


class ClientActorSPStub(object):
    """----------------------------------------------------------------------------------------------------------

    """

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.RunTrial = channel.stream_stream(
                '/cogmentAPI.ClientActorSP/RunTrial',
                request_serializer=cogment_dot_api_dot_common__pb2.ActorRunTrialOutput.SerializeToString,
                response_deserializer=cogment_dot_api_dot_common__pb2.ActorRunTrialInput.FromString,
                )
        self.Version = channel.unary_unary(
                '/cogmentAPI.ClientActorSP/Version',
                request_serializer=cogment_dot_api_dot_common__pb2.VersionRequest.SerializeToString,
                response_deserializer=cogment_dot_api_dot_common__pb2.VersionInfo.FromString,
                )


class ClientActorSPServicer(object):
    """----------------------------------------------------------------------------------------------------------

    """

    def RunTrial(self, request_iterator, context):
        """Expected metadata:
        - trial-id: The id of the trial
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Version(self, request, context):
        """Expected metadata: None
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_ClientActorSPServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'RunTrial': grpc.stream_stream_rpc_method_handler(
                    servicer.RunTrial,
                    request_deserializer=cogment_dot_api_dot_common__pb2.ActorRunTrialOutput.FromString,
                    response_serializer=cogment_dot_api_dot_common__pb2.ActorRunTrialInput.SerializeToString,
            ),
            'Version': grpc.unary_unary_rpc_method_handler(
                    servicer.Version,
                    request_deserializer=cogment_dot_api_dot_common__pb2.VersionRequest.FromString,
                    response_serializer=cogment_dot_api_dot_common__pb2.VersionInfo.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'cogmentAPI.ClientActorSP', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))


 # This class is part of an EXPERIMENTAL API.
class ClientActorSP(object):
    """----------------------------------------------------------------------------------------------------------

    """

    @staticmethod
    def RunTrial(request_iterator,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.stream_stream(request_iterator, target, '/cogmentAPI.ClientActorSP/RunTrial',
            cogment_dot_api_dot_common__pb2.ActorRunTrialOutput.SerializeToString,
            cogment_dot_api_dot_common__pb2.ActorRunTrialInput.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def Version(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/cogmentAPI.ClientActorSP/Version',
            cogment_dot_api_dot_common__pb2.VersionRequest.SerializeToString,
            cogment_dot_api_dot_common__pb2.VersionInfo.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)



# Copyright 2021 AI Redefined Inc. <dev+cogment@ai-r.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import grpc
import grpc.aio  # type: ignore

import cogment.api.common_pb2 as common_api
import cogment.api.orchestrator_pb2 as orchestrator_api
from cogment.errors import CogmentError
from cogment.session import ActorInfo
from cogment.utils import logger

import asyncio
from enum import Enum


class TrialState(Enum):
    """Enum class for the different states of a trial."""

    UNKNOWN = common_api.TrialState.UNKNOWN
    INITIALIZING = common_api.TrialState.INITIALIZING
    PENDING = common_api.TrialState.PENDING
    RUNNING = common_api.TrialState.RUNNING
    TERMINATING = common_api.TrialState.TERMINATING
    ENDED = common_api.TrialState.ENDED


class TrialInfo:
    """Class representing the details of a trial."""

    def __init__(self, trial_id, api_state):
        self.trial_id = trial_id
        self.state = TrialState(api_state)
        self.env_name = None
        self.tick_id = None
        self.duration = None

    def __str__(self):
        result = f"TrialInfo: trial_id = {self.trial_id}, env_name = {self.env_name}, state = {self.state}"
        result += f", tick_id = {self.tick_id}, duration = {self.duration}"
        return result


class Controller:
    """Class representing a Cogment controller associated with an Orchestrator."""

    def __init__(self, stub, user_id):
        self._lifecycle_stub = stub
        self._user_id = user_id

    def __str__(self):
        result = f"Controller: user id = {self._user_id}"
        return result

    async def get_actors(self, trial_id):
        req = orchestrator_api.TrialInfoRequest()
        req.get_actor_list = True
        metadata = [("trial-id", trial_id)]
        rep = await self._lifecycle_stub.GetTrialInfo(request=req, metadata=metadata)
        if len(rep.trial) == 0:
            raise CogmentError(f"Unknown trial [{trial_id}]")
        elif len(rep.trial) > 1:
            raise CogmentError(f"Unexpected response from orchestrator [{rep}] for [{trial_id}]")

        result = [
            ActorInfo(actor.name, actor.actor_class)
            for actor in rep.trial[0].actors_in_trial
        ]
        return result

    async def start_trial(self, trial_config=None, trial_id_requested=None, trial_params=None):
        req = orchestrator_api.TrialStartRequest()
        req.user_id = self._user_id

        if trial_config is not None:
            if trial_params is not None:
                raise CogmentError(f"Cannot provide both a start config and start parameters")
            req.config.content = trial_config.SerializeToString()
        elif trial_params is not None:
            req.params.CopyFrom(trial_params._raw_params)
            if trial_params._raw_params.HasField("trial_config"):
                logger.warning(f"Trial config from trial parameters will be ignored")

        if trial_id_requested is not None:
            req.trial_id_requested = trial_id_requested

        logger.debug(f"Requesting start of a trial with [{req}] ...")
        rep = await self._lifecycle_stub.StartTrial(req)

        if rep.trial_id:
            logger.debug(f"Trial [{rep.trial_id}] started")
        else:
            logger.warning(f"Requested trial id [{trial_id_requested}] could not be used")

        return rep.trial_id

    async def terminate_trial(self, trial_ids, hard=False):
        req = orchestrator_api.TerminateTrialRequest()
        req.hard_termination = hard
        if type(trial_ids) == str:
            logger.deprecated("Using Controller.terminate_trial() with a string trial ID is deprecated.  Use a list.")
            metadata = [("trial-id", trial_ids)]
        else:
            metadata = []
            for id in trial_ids:
                metadata.append(("trial-id", id))

        logger.debug(f"Requesting end of trial [{trial_ids}] (hard termination: [{hard}])")
        await self._lifecycle_stub.TerminateTrial(request=req, metadata=metadata)

    async def get_remote_versions(self):
        req = common_api.VersionRequest()
        info = await self._lifecycle_stub.Version(request=req)
        result = {}
        for ver in info.versions:
            result[ver.name] = ver.version
        return result

    async def get_trial_info(self, trial_ids):
        req = orchestrator_api.TrialInfoRequest()
        if trial_ids is None:
            logger.deprecated("Using Controller.get_trial_info() with a null trial ID is deprecated.  Use a list.")
            metadata = []
        elif type(trial_ids) == str:
            logger.deprecated("Using Controller.get_trial_info() with a string trial ID is deprecated.  Use a list.")
            metadata = [("trial-id", trial_ids)]
        else:
            metadata = []
            for id in trial_ids:
                metadata.append(("trial-id", id))
        rep = await self._lifecycle_stub.GetTrialInfo(request=req, metadata=metadata)

        result = []
        for reply in rep.trial:
            info_ex = TrialInfo(reply.trial_id, reply.state)
            info_ex.env_name = reply.env_name
            info_ex.tick_id = reply.tick_id
            info_ex.duration = reply.trial_duration

            result.append(info_ex)

        return result

    async def watch_trials(self, trial_state_filters=[]):
        request = orchestrator_api.TrialListRequest()
        for filter in trial_state_filters:
            if type(filter) != TrialState:
                raise CogmentError(f"Unknown filter type [{type(filter)}]: must of type 'cogment.TrialState'")
            request.filter.append(filter.value)

        reply_itor = self._lifecycle_stub.WatchTrials(request=request)
        if not reply_itor:
            raise CogmentError(f"'watch_trials' failed to connect")

        try:
            async for reply in reply_itor:
                info = TrialInfo(reply.trial_id, reply.state)
                keep_looping = yield info
                if keep_looping is not None and not bool(keep_looping):
                    break

        except grpc.aio.AioRpcError as exc:
            logger.debug(f"gRPC failed status details: [{exc.debug_error_string()}]")
            if exc.code() == grpc.StatusCode.UNAVAILABLE:
                logger.error(f"Watch_trials Orchestrator communication lost: [{exc.details()}]")
            else:
                logger.exception("watch_trials -- Unexpected aio failure")
                raise

        except GeneratorExit:
            raise

        except asyncio.CancelledError as exc:
            logger.debug(f"watch_trial coroutine cancelled while waiting for trial info: [{exc}]")

        except Exception:
            logger.exception("watch_trials")
            raise



# Copyright 2021 AI Redefined Inc. <dev+cogment@ai-r.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import grpc
import grpc.aio  # type: ignore

import cogment.api.common_pb2 as common_api

from cogment.actor import ActorSession
from cogment.utils import logger, INIT_TIMEOUT
from cogment.errors import CogmentError
from cogment.session import RecvEvent, RecvObservation, RecvMessage, RecvReward, EventType, _InitAck, _EndingAck
from cogment.trial import Trial

import asyncio


class _EndQueue:
    """Internal class signaling the end of data queueing."""
    pass


def _impl_can_serve_actor_class(impl, actor_class):
    if impl.actor_classes:
        for ac in impl.actor_classes:
            if ac == actor_class.name:
                return True
        return False
    else:
        return True


def _process_normal_data(data, session):
    if session._trial.ending:
        recv_event = RecvEvent(EventType.ENDING)
    else:
        recv_event = RecvEvent(EventType.ACTIVE)

    if data.HasField("observation"):
        logger.trace(f"Trial [{session._trial.id}] - Actor [{session.name}] received an observation")

        if session._trial.ending and session._auto_ack:
            session._post_outgoing_data(_EndingAck())

        tick_id = data.observation.tick_id
        session._trial.tick_id = tick_id

        obs_space = session._actor_class.observation_space()
        obs_space.ParseFromString(data.observation.content)

        recv_event.observation = RecvObservation(data.observation, obs_space)
        session._post_incoming_event((tick_id, recv_event))

    elif data.HasField("reward"):
        logger.trace(f"Trial [{session._trial.id}] - Actor [{session.name}] received reward")

        recv_event.rewards = [RecvReward(data.reward)]
        session._post_incoming_event((-1, recv_event))

    elif data.HasField("message"):
        logger.trace(f"Trial [{session._trial.id}] - Actor [{session.name}] received message")

        recv_event.messages = [RecvMessage(data.message)]
        session._post_incoming_event((-1, recv_event))

    elif data.HasField("details"):
        logger.warning(f"Trial [{session._trial.id}] - Actor [{session.name}] "
                       f"received unexpected detail data [{data.details}]")

    else:
        logger.error(f"Trial [{session._trial.id}] - Actor [{session.name}] "
                     f"received unexpected data [{data.WhichOneof('data')}]")


async def _process_incoming(reply_itor, req_queue, session):
    try:
        async for data in reply_itor:
            if data == grpc.aio.EOF:
                logger.info(f"Trial [{session._trial.id}] - Actor [{session.name}]: "
                            f"The orchestrator disconnected the actor")
                break
            logger.trace(f"Trial [{session._trial.id}] - Actor [{session.name}]: "
                         f"Received data [{data.state}] [{data.WhichOneof('data')}]")

            if data.state == common_api.CommunicationState.NORMAL:
                _process_normal_data(data, session)

            elif data.state == common_api.CommunicationState.HEARTBEAT:
                logger.trace(f"Trial [{session._trial.id}] - Actor [{session.name}] "
                             f"received 'HEARTBEAT' and responding in kind")
                reply = common_api.ActorRunTrialOutput()
                reply.state = data.state
                await req_queue.put(reply)

            elif data.state == common_api.CommunicationState.LAST:
                logger.debug(f"Trial [{session._trial.id}] - Actor [{session.name}] received 'LAST' state")
                session._trial.ending = True

            elif data.state == common_api.CommunicationState.LAST_ACK:
                logger.error(f"Trial [{session._trial.id}] - Actor [{session.name}] "
                             f"received an unexpected 'LAST_ACK'")
                # TODO: Should we `return` or raise instead of continuing?

            elif data.state == common_api.CommunicationState.END:
                if session._trial.ending:
                    if data.HasField("details"):
                        logger.info(f"Trial [{session._trial.id}] - Actor [{session.name}]: "
                                    f"Trial ended with explanation [{data.details}]")
                    else:
                        logger.debug(f"Trial [{session._trial.id}] - Actor [{session.name}]: Trial ended")
                    session._post_incoming_event((-1, RecvEvent(EventType.FINAL)))
                else:
                    if data.HasField("details"):
                        details = data.details
                    else:
                        details = ""
                    logger.warning(f"Trial [{session._trial.id}] - Actor [{session.name}]: "
                                   f"Trial ended forcefully [{details}]")

                session._trial.ended = True
                session._exit_queues()
                break

            else:
                logger.error(f"Trial [{session._trial.id}] - Actor [{session.name}] "
                             f"received an invalid state [{data.state}]")

    except asyncio.CancelledError as exc:
        logger.debug(f"Trial [{session._trial.id}] - Actor [{session.name}] coroutine cancelled: [{exc}]")
        raise

    except grpc.aio.AioRpcError as exc:
        logger.debug(f"gRPC failed status details: [{exc.debug_error_string()}]")
        if exc.code() == grpc.StatusCode.UNAVAILABLE:
            logger.error(f"Trial [{session._trial.id}] - Actor [{session.name}] "
                         f"Orchestrator communication lost [{exc.details()}]")
            session._exit_queues()
        else:
            logger.exception("_process_incoming -- Unexpected aio failure")
            raise

    except Exception:
        logger.exception("_process_incoming")
        raise


async def _process_outgoing(data_queue, session):
    try:
        async for data in session._retrieve_outgoing_data():
            package = common_api.ActorRunTrialOutput()
            package.state = common_api.CommunicationState.NORMAL

            # Using strict comparison: there is no reason to receive derived classes here
            if type(data) == common_api.Action:
                logger.trace(f"Trial [{session._trial.id}] - Actor [{session.name}]: Sending action")
                package.action.CopyFrom(data)

            elif type(data) == common_api.Reward:
                logger.trace(f"Trial [{session._trial.id}] - Actor [{session.name}]: Sending reward")
                package.reward.CopyFrom(data)

            elif type(data) == common_api.Message:
                logger.trace(f"Trial [{session._trial.id}] - Actor [{session.name}]: Sending message")
                package.message.CopyFrom(data)

            elif type(data) == _InitAck:
                # The init output was done before we could get here
                continue

            elif type(data) == _EndingAck:
                package.state = common_api.CommunicationState.LAST_ACK
                logger.debug(f"Trial [{session._trial.id}] - Actor [{session.name}]: Sending 'LAST_ACK'")
                await data_queue.put(package)
                break

            else:
                logger.error(f"Trial [{session._trial.id}] - Actor [{session.name}]: "
                             f"Unknown data type to send [{type(data)}]")
                continue

            await data_queue.put(package)

    except asyncio.CancelledError as exc:
        logger.debug(f"Trial [{session._trial.id}] - Actor [{session.name}] process outgoing cancelled: [{exc}]")
        raise

    except Exception:
        logger.exception("_process_outgoing")
        raise


class _GrpcWriter:
    """Internal class encapsulating an async queue to serve as a grpc writing iterator."""

    def __init__(self, data_queue):
        self._queue = data_queue

    def __aiter__(self):
        return self

    async def __anext__(self):
        try:
            data = await self._queue.get()
            self._queue.task_done()
            if (type(data) != _EndQueue):
                return data

        except asyncio.CancelledError as exc:
            logger.debug(f"Client '_GrpcWriter' coroutine cancelled: [{exc}]")

        except GeneratorExit:
            raise

        except Exception:
            logger.exception("_GrpcWriter::anext")
            raise

        raise StopAsyncIteration


class ClientServicer:
    """Internal client actor servicer class."""

    def __init__(self, cog_settings, stub):
        self.trial_id = None
        self._cog_settings = cog_settings
        self._request_queue = None
        self._reply_itor = None
        self._actor_stub = stub

    async def _get_init_data(self):
        logger.debug(f"Trial [{self.trial_id}] - Processing init for client actor")

        last_received = False
        async for reply in self._reply_itor:
            logger.debug(f"Trial [{self.trial_id}] - Read initial reply: {reply}")

            if reply == grpc.aio.EOF:
                logger.info(f"Trial [{self.trial_id}] - Orchestrator disconnected client actor before start")
                return None

            if reply.state == common_api.CommunicationState.NORMAL:
                if last_received:
                    raise CogmentError(f"Trial [{self.trial_id}] - Received init data after 'LAST'")

                if reply.HasField("init_input"):
                    return reply.init_input
                elif reply.HasField("details"):
                    logger.warning(f"Trial [{self.trial_id}] - Received unexpected detail data "
                                   f"[{reply.details}] before start")
                else:
                    data_set = reply.WhichOneof("data")
                    raise CogmentError(f"Trial [{self.trial_id}] - Received unexpected init data [{data_set}]")

            elif reply.state == common_api.CommunicationState.HEARTBEAT:
                req = common_api.ActorRunTrialOutput()
                req.state = reply.state
                await self._request_queue.put(req)

            elif reply.state == common_api.CommunicationState.LAST:
                if last_received:
                    raise CogmentError(f"Trial [{self.trial_id}] - Before start, received unexpected 'LAST' "
                                       f"when waiting for 'END'")

                logger.debug(f"Trial [{self.trial_id}] - Ending before init data")
                req = common_api.ActorRunTrialOutput()
                req.state = common_api.CommunicationState.LAST_ACK
                await self._request_queue.put(req)
                last_received = True

            elif reply.state == common_api.CommunicationState.LAST_ACK:
                raise CogmentError(f"Trial [{self.trial_id}] - Before start, received an unexpected 'LAST_ACK'")

            elif reply.state == common_api.CommunicationState.END:
                if reply.HasField("details"):
                    logger.info(f"Trial [{self.trial_id}] - Ended before start [{reply.details}]")
                else:
                    logger.debug(f"Trial [{self.trial_id}] - Ended before start")
                return None

            else:
                raise CogmentError(f"Trial [{self.trial_id}] - Before start, received an invalid state "
                                   f"[{reply.state}] [{reply}]")

    def _start_session(self, impl, init_data):
        actor_name = init_data.actor_name
        if not actor_name:
            raise CogmentError(f"Trial [{self.trial_id}] - Empty actor name for client actor")

        actor_class = self._cog_settings.actor_classes.get(init_data.actor_class)
        if actor_class is None:
            raise CogmentError(f"Trial [{self.trial_id}] - "
                               f"Unknown class [{init_data.actor_class}] for client actor [{actor_name}]")

        config = None
        if init_data.HasField("config"):
            if actor_class.config_type is None:
                raise CogmentError(f"Trial [{self.trial_id}] - Client actor [{actor_name}] "
                                   f"received config data of unknown type (was it defined in cogment.yaml?)")
            config = actor_class.config_type()
            config.ParseFromString(init_data.config.content)

        trial = Trial(self.trial_id, [], self._cog_settings)
        new_session = ActorSession(impl, actor_class, trial, actor_name, init_data.impl_name,
                                   init_data.env_name, config)

        logger.debug(f"Trial [{self.trial_id}] - impl [{init_data.impl_name}] for actor [{actor_name}] started")

        return new_session

    async def run_session(self, impl, init_data):
        if self._request_queue is None:
            raise CogmentError(f"ClientServicer has not joined")

        send_task = None
        process_task = None

        session = self._start_session(impl, init_data)

        try:
            send_task = asyncio.create_task(_process_outgoing(self._request_queue, session))
            process_task = asyncio.create_task(_process_incoming(self._reply_itor, self._request_queue, session))
            user_task = session._start_user_task()

            logger.debug(f"Trial [{session._trial.id}] - Actor [{session.name}] session started")
            normal_return = await user_task

            if normal_return:
                if not session._last_event_delivered:
                    logger.warning(f"Trial [{session._trial.id}] - Actor [{session.name}] "
                                   f"user implementation returned before required")
                else:
                    logger.debug(f"Trial [{session._trial.id}] - Actor [{session.name}] "
                                 f"user implementation returned")
            else:
                logger.debug(f"Trial [{session._trial.id}] - Actor [{session.name}] "
                             f"user implementation was cancelled")

        except asyncio.CancelledError as exc:
            logger.debug(f"Trial [{session._trial.id}] - Actor [{session.name}] "
                         f"user implementation was cancelled")

        except Exception:
            logger.exception("run_session")
            raise

        finally:
            if self._request_queue is not None:
                await self._request_queue.put(_EndQueue())
            if send_task is not None:
                send_task.cancel()
            if process_task is not None:
                process_task.cancel()

    async def join_trial(self, trial_id, actor_name, actor_class):
        if self._request_queue is not None:
            raise CogmentError(f"ClientServicer has already joined")

        req = common_api.ActorRunTrialOutput()
        req.state = common_api.CommunicationState.NORMAL

        init = common_api.ActorInitialOutput()
        if actor_name is not None:
            init.actor_name = actor_name
        elif actor_class is not None:
            init.actor_class = actor_class
        else:
            raise CogmentError(f"Only actor_name or actor_class must be specified, not both.")
        req.init_output.CopyFrom(init)

        self._request_queue = asyncio.Queue()
        await self._request_queue.put(req)

        self.trial_id = trial_id

        req_itor = _GrpcWriter(self._request_queue)
        metadata = (("trial-id", trial_id),)
        self._reply_itor = self._actor_stub.RunTrial(request_iterator=req_itor, metadata=metadata)
        if not self._reply_itor:
            raise CogmentError(f"Failed to connect to join trial")

        try:
            return await asyncio.wait_for(self._get_init_data(), INIT_TIMEOUT)

        except CogmentError:
            raise

        except asyncio.TimeoutError:
            logger.error(f"Trial [{trial_id}] - Join failed to receive init data from Orchestrator")

        except asyncio.CancelledError as exc:
            logger.error(f"Trial [{trial_id}] - Join coroutine cancelled before start: [{exc}]")

        except grpc.aio.AioRpcError as exc:
            logger.debug(f"gRPC failed status details: [{exc.debug_error_string()}]")
            if exc.code() == grpc.StatusCode.UNAVAILABLE:
                logger.error(f"Trial [{trial_id}] - Actor [{actor_name}] "
                             f"Orchestrator communication lost in init [{exc.details()}]")
            else:
                logger.exception("join_trial -- Unexpected aio failure")
                raise

        except Exception:
            logger.exception("join_trial")
            raise

        return None



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: cogment/api/hooks.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from cogment.api import common_pb2 as cogment_dot_api_dot_common__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x17\x63ogment/api/hooks.proto\x12\ncogmentAPI\x1a\x18\x63ogment/api/common.proto\"9\n\x0ePreTrialParams\x12\'\n\x06params\x18\x01 \x01(\x0b\x32\x17.cogmentAPI.TrialParams2\x98\x01\n\x0cTrialHooksSP\x12\x46\n\nOnPreTrial\x12\x1a.cogmentAPI.PreTrialParams\x1a\x1a.cogmentAPI.PreTrialParams\"\x00\x12@\n\x07Version\x12\x1a.cogmentAPI.VersionRequest\x1a\x17.cogmentAPI.VersionInfo\"\x00\x62\x06proto3')



_PRETRIALPARAMS = DESCRIPTOR.message_types_by_name['PreTrialParams']
PreTrialParams = _reflection.GeneratedProtocolMessageType('PreTrialParams', (_message.Message,), {
  'DESCRIPTOR' : _PRETRIALPARAMS,
  '__module__' : 'cogment.api.hooks_pb2'
  # @@protoc_insertion_point(class_scope:cogmentAPI.PreTrialParams)
  })
_sym_db.RegisterMessage(PreTrialParams)

_TRIALHOOKSSP = DESCRIPTOR.services_by_name['TrialHooksSP']
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _PRETRIALPARAMS._serialized_start=65
  _PRETRIALPARAMS._serialized_end=122
  _TRIALHOOKSSP._serialized_start=125
  _TRIALHOOKSSP._serialized_end=277
# @@protoc_insertion_point(module_scope)



# Copyright 2021 AI Redefined Inc. <dev+cogment@ai-r.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import cogment.api.hooks_pb2_grpc as hooks_grpc_api
import cogment.api.hooks_pb2 as hooks_api

from cogment.trial import Trial
from cogment.prehook import PrehookSession
from cogment.parameters import TrialParameters
from cogment.errors import CogmentError
from cogment.utils import list_versions, logger

import asyncio


class PrehookServicer(hooks_grpc_api.TrialHooksSPServicer):
    """Internal pre-trial hook servicer class."""

    def __init__(self, impl, cog_settings, prometheus_registry=None):
        self._impl = impl
        self._cog_settings = cog_settings

        logger.info("Pre-trial hook service started")

    # Not all attributes can be tested for presence (i.e. non-optional, non-message ones)
    def _decode(self, session, proto_params):
        if proto_params.HasField("trial_config"):
            session.trial_config = self._cog_settings.trial.config_type()
            session.trial_config.ParseFromString(proto_params.trial_config.content)
        else:
            session.trial_config = None

        session.trial_max_steps = proto_params.max_steps
        session.trial_max_inactivity = proto_params.max_inactivity

        if proto_params.HasField("datalog"):
            session.datalog_endpoint = proto_params.datalog.endpoint
            session.datalog_exclude = [field for field in proto_params.datalog.exclude_fields]
        else:
            session.datalog_endpoint = None
            session.datalog_exclude = []

        if proto_params.HasField("environment"):
            if proto_params.environment.HasField("config"):
                session.environment_config = self._cog_settings.environment.config_type()
                session.environment_config.ParseFromString(proto_params.environment.config.content)
            else:
                session.environment_config = None
            session.environment_endpoint = proto_params.environment.endpoint
            session.environment_name = proto_params.environment.name
            session.environment_implementation = proto_params.environment.implementation
        else:
            session.environment_config = None
            session.environment_endpoint = None
            session.environment_name = None
            session.environment_implementation = None

        session.actors = []
        for actor in proto_params.actors:
            if actor.HasField("config"):
                a_c = self._cog_settings.actor_classes.__getattribute__(actor.actor_class)
                actor_config = a_c.config_type()
                actor_config.ParseFromString(actor.config.content)
            else:
                actor_config = None
            actor_name = actor.name
            actor_class = actor.actor_class
            actor_endpoint = actor.endpoint
            actor_implementation = actor.implementation

            actor_data = {
                "name": actor_name,
                "actor_class": actor_class,
                "endpoint": actor_endpoint,
                "implementation": actor_implementation,
                "config": actor_config
            }
            session.actors.append(actor_data)

    def _recode(self, proto_params, session):
        if hasattr(session, "trial_config") and session.trial_config is not None:
            proto_params.trial_config.content = session.trial_config.SerializeToString()
        if hasattr(session, "trial_max_steps") and session.trial_max_steps is not None:
            proto_params.max_steps = session.trial_max_steps
        if hasattr(session, "trial_max_inactivity") and session.trial_max_inactivity is not None:
            proto_params.max_inactivity = session.trial_max_inactivity

        if hasattr(session, "datalog_endpoint") and session.datalog_endpoint is not None:
            proto_params.datalog.endpoint = session.datalog_endpoint
            if hasattr(session, "datalog_exclude") and session.datalog_exclude is not None:
                proto_params.datalog.exclude_fields.extend(session.datalog_exclude)

        if hasattr(session, "environment_config") and session.environment_config is not None:
            proto_params.environment.config.content = session.environment_config.SerializeToString()
        if hasattr(session, "environment_endpoint") and session.environment_endpoint is not None:
            proto_params.environment.endpoint = session.environment_endpoint
        if hasattr(session, "environment_name") and session.environment_name is not None:
            proto_params.environment.name = session.environment_name
        if hasattr(session, "environment_implementation") and session.environment_implementation is not None:
            proto_params.environment.implementation = session.environment_implementation

        if hasattr(session, "actors") and session.actors is not None:
            for actor_data in session.actors:
                actor_pb = proto_params.actors.add()
                if "name" in actor_data:
                    actor_pb.name = actor_data["name"]
                if "actor_class" in actor_data:
                    actor_pb.actor_class = actor_data["actor_class"]
                if "endpoint" in actor_data:
                    actor_pb.endpoint = actor_data["endpoint"]
                if "implementation" in actor_data:
                    actor_pb.implementation = actor_data["implementation"]
                if "config" in actor_data and actor_data["config"] is not None:
                    actor_pb.config.content = actor_data["config"].SerializeToString()

    # Override
    async def OnPreTrial(self, request, context):
        if not self._impl:
            raise CogmentError("No implementation registered")

        try:
            metadata = dict(context.invocation_metadata())
            logger.debug(f"Received metadata: [{metadata}]")
            trial_id = metadata["trial-id"]
            user_id = metadata["user-id"]

            trial = Trial(trial_id, [], self._cog_settings)

            trial_parameters = TrialParameters(None)
            trial_parameters._set(self._cog_settings, request.params)

            session = PrehookSession(trial, user_id, trial_parameters)
            self._decode(session, request.params)

            try:
                await self._impl(session)
                session.validate()

            except asyncio.CancelledError as exc:
                logger.debug(f"Pre-trial hook implementation coroutine cancelled: [{exc}]")
                return False

            except Exception:
                logger.exception(f"An exception occured in user prehook implementation:")
                raise

            reply = hooks_api.PreTrialParams()

            if session._trial_parameters_use:
                reply.params.CopyFrom(session._trial_parameters._raw_params)
            else:
                logger.deprecated(f"Use of pre-trial hook session in a deprecated way")
                self._recode(reply.params, session)

            return reply

        except Exception:
            logger.exception("OnPreTrial")
            raise

    # Override
    async def Version(self, request, context):
        try:
            return list_versions()

        except Exception:
            logger.exception("Version")
            raise



# Copyright 2021 AI Redefined Inc. <dev+cogment@ai-r.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import grpc
import grpc.aio  # type: ignore

import cogment.api.common_pb2 as common_api
import cogment.api.trial_datastore_pb2 as datastore_api

from cogment.control import TrialState
from cogment.parameters import TrialParameters
from cogment.errors import CogmentError
from cogment.utils import logger

import asyncio
import enum
import datetime

_REV_NANO = 1.0 / 1_000_000_000


class DatastoreFields(enum.Enum):
    """Enum class for the different fields of the actor data that can be retrieved."""

    UNKNOWN = datastore_api.StoredTrialSampleField.STORED_TRIAL_SAMPLE_FIELD_UNKNOWN
    OBSERVATION = datastore_api.StoredTrialSampleField.STORED_TRIAL_SAMPLE_FIELD_OBSERVATION
    ACTION = datastore_api.StoredTrialSampleField.STORED_TRIAL_SAMPLE_FIELD_ACTION
    REWARD = datastore_api.StoredTrialSampleField.STORED_TRIAL_SAMPLE_FIELD_REWARD
    RECEIVED_REWARDS = datastore_api.StoredTrialSampleField.STORED_TRIAL_SAMPLE_FIELD_RECEIVED_REWARDS
    SENT_REWARDS = datastore_api.StoredTrialSampleField.STORED_TRIAL_SAMPLE_FIELD_SENT_REWARDS
    RECEIVED_MESSAGES = datastore_api.StoredTrialSampleField.STORED_TRIAL_SAMPLE_FIELD_RECEIVED_MESSAGES
    SENT_MESSAGES = datastore_api.StoredTrialSampleField.STORED_TRIAL_SAMPLE_FIELD_SENT_MESSAGES


class DatastoreTrialInfo:
    """Class representing trial info."""

    def __init__(self, cog_settings, info: datastore_api.StoredTrialInfo):
        self.trial_id = info.trial_id
        self.trial_state = TrialState(info.last_state)
        self.user_id = info.user_id
        self.sample_count = info.samples_count

        self.parameters = TrialParameters(None)
        self.parameters._set(cog_settings, info.params)

    def __str__(self):
        result = f"DatastoreTrialInfo:"
        result += f" trial_id = {self.trial_id}, trial_state = {self.trial_state}"
        result += f", user_id = {self.user_id}, sample_count = {self.sample_count}"
        result += f", parameters = {self.parameters}"
        return result


class DatastoreReward:
    """Class representing an individual reward sent during this sample."""

    def __init__(self, reward, payloads, parameters: TrialParameters):
        self._raw_reward = reward
        self._payloads = payloads
        self._parameters = parameters

        self.value = self._raw_reward.reward
        self.confidence = self._raw_reward.confidence

    def __str__(self):
        result = f"DatastoreReward:"
        result += f" value = {self.value}, confidence = {self.confidence}"
        result += f", sender = {self.sender}, receiver = {self.receiver}, user_data = {self.user_data}"
        return result

    @property
    def sender(self):
        """Name of actor that sent the reward"""
        actor_index = self._raw_reward.sender
        return self._parameters.actors[actor_index].name

    @property
    def receiver(self):
        """Name of actor that received the reward"""
        actor_index = self._raw_reward.receiver
        return self._parameters.actors[actor_index].name

    @property
    def user_data(self):
        """User data sent with the reward"""
        if self._raw_reward.HasField("user_data"):
            data_index = self._raw_reward.user_data
            data_content = self._payloads[data_index]
            any = common_api.RewardSource().user_data  # Easier than instantiating 'google.protobuf.any_pb2.Any'
            any.ParseFromString(data_content)

            return any
        else:
            return None


class DatastoreMessage:
    """Class representing an individual message sent during this sample."""

    def __init__(self, message, payloads, parameters: TrialParameters):
        self._raw_message = message
        self._payloads = payloads
        self._parameters = parameters

    def __str__(self):
        result = f"DatastoreMessage:"
        result += f", sender = {self.sender}, receiver = {self.receiver}, payload = {self.payload}"
        return result

    @property
    def sender(self):
        """Name of actor that sent the message"""
        actor_index = self._raw_message.sender
        return self._parameters.actors[actor_index].name

    @property
    def receiver(self):
        """Name of actor that received the message"""
        actor_index = self._raw_message.receiver
        return self._parameters.actors[actor_index].name

    @property
    def payload(self):
        """Payload sent with the reward"""
        if self._raw_message.HasField("payload"):
            payload_index = self._raw_message.payload
            payload_content = self._payloads[payload_index]
            any = common_api.Message().payload  # Easier than instantiating 'google.protobuf.any_pb2.Any'
            any.ParseFromString(payload_content)

            return any
        else:
            return None


class DatastoreActorData:
    """Class representing the data for an actor in a sample."""

    def __init__(self, sample, payloads, parameters: TrialParameters):
        self._raw_sample = sample
        self._payloads = payloads
        self._parameters = parameters

    def __str__(self):
        result = f"DatastoreActorData:"
        result += f" name = {self.name}, reward = {self.reward}"
        result += f", observation = {self.observation}, action = {self.action}"
        result += f", nb received rewards = {len(self._raw_sample.received_rewards)}"
        result += f", nb sent rewards = {len(self._raw_sample.sent_rewards)}"
        result += f", nb received messages = {len(self._raw_sample.received_messages)}"
        result += f", nb sent messages = {len(self._raw_sample.sent_messages)}"
        return result

    @property
    def name(self):
        """Name of the actor"""
        actor_index = self._raw_sample.actor
        return self._parameters.actors[actor_index].name

    @property
    def observation(self):
        """Observation space to the actor"""
        if self._raw_sample.HasField("observation"):
            actor_index = self._raw_sample.actor
            obs_space = self._parameters.actors[actor_index]._actor_class.observation_space()

            obs_index = self._raw_sample.observation
            obs_content = self._payloads[obs_index]
            obs_space.ParseFromString(obs_content)
            return obs_space
        else:
            return None

    @property
    def action(self):
        """Action space from the actor"""
        if self._raw_sample.HasField("action"):
            actor_index = self._raw_sample.actor
            action_space = self._parameters.actors[actor_index]._actor_class.action_space()

            action_index = self._raw_sample.action
            action_content = self._payloads[action_index]
            action_space.ParseFromString(action_content)
            return action_space
        else:
            return None

    @property
    def reward(self):
        """Aggregated reward received by the actor"""
        if self._raw_sample.HasField("reward"):
            return self._raw_sample.reward
        else:
            return None

    def all_received_rewards(self):
        for rew in self._raw_sample.received_rewards:
            yield DatastoreReward(rew, self._payloads, self._parameters)

    def all_sent_rewards(self):
        for rew in self._raw_sample.sent_rewards:
            yield DatastoreReward(rew, self._payloads, self._parameters)

    def all_received_messages(self):
        for msg in self._raw_sample.received_messages:
            yield DatastoreMessage(msg, self._payloads, self._parameters)

    def all_sent_messages(self):
        for msg in self._raw_sample.sent_messages:
            yield DatastoreMessage(msg, self._payloads, self._parameters)


class DatastoreSample:
    """Class representing a trial sample from the trial datastore service."""

    def __init__(self, sample, params: TrialParameters):
        self.trial_id = sample.trial_id
        self.trial_state = TrialState(sample.state)
        self.tick_id = sample.tick_id
        self.timestamp = sample.timestamp

        self._raw_sample = sample
        self._parameters = params

        self.actors_data = {}
        for smpl in self._raw_sample.actor_samples:
            data = DatastoreActorData(smpl, self._raw_sample.payloads, self._parameters)
            self.actors_data[data.name] = data
        if len(self.actors_data) != len(self._raw_sample.actor_samples):
            raise CogmentError(f"Duplicate actor names in datastore sample")

    def __str__(self):
        utc_time = datetime.datetime.utcfromtimestamp(self.timestamp * _REV_NANO)
        result = f"DatastoreSample:"
        result += f" trial_id = {self.trial_id}, trial_state = {self.trial_state}"
        result += f", tick_id = {self.tick_id}, timestamp = {self.timestamp} UTC[{utc_time}]"
        result += f", nb actors = {len(self.actors_data)}"
        return result


class Datastore:
    """Class representing the session of a datalog for a trial."""

    def __init__(self, stub, cog_settings):
        self._datastore_stub = stub
        self._cog_settings = cog_settings
        self._timeout = 0

    def __str__(self):
        result = f"Datastore"
        return result

    async def all_trials(self, bundle_size=1):
        request = datastore_api.RetrieveTrialsRequest()
        request.timeout = 0
        request.trials_count = bundle_size
        request.trial_handle = ""

        while True:
            reply = await self._datastore_stub.RetrieveTrials(request)

            for rep_info in reply.trial_infos:
                info = DatastoreTrialInfo(self._cog_settings, rep_info)
                yield info

            if len(reply.trial_infos) < bundle_size:
                break
            if reply.next_trial_handle:
                request.trial_handle = reply.next_trial_handle
            else:
                break

    async def get_trials(self, ids):
        request = datastore_api.RetrieveTrialsRequest()
        request.timeout = self._timeout
        for id in ids:
            request.trial_ids.append(id)

        reply = await self._datastore_stub.RetrieveTrials(request)

        trial_infos = []
        for rep_info in reply.trial_infos:
            info = DatastoreTrialInfo(self._cog_settings, rep_info)
            trial_infos.append(info)

        return trial_infos

    async def delete_trials(self, ids):
        if not ids:
            raise CogmentError("At least one trial ID must be provided to delete")

        request = datastore_api.DeleteTrialsRequest()
        for id in ids:
            request.trial_ids.append(id)

        await self._datastore_stub.DeleteTrials(request)

    async def all_samples(self, trial_infos, actor_names=[], actor_classes=[], actor_implementations=[], fields=[]):
        if not trial_infos:
            raise CogmentError("At least one trial info must be provided to retrieve samples")

        request = datastore_api.RetrieveSamplesRequest()

        params = {}
        for info in trial_infos:
            request.trial_ids.append(info.trial_id)
            params[info.trial_id] = info.parameters
        for name in actor_names:
            request.actor_names.append(name)
        for cls in actor_classes:
            request.actor_classes.append(cls)
        for impl in actor_implementations:
            request.actor_implementations.append(impl)
        for enum_field in fields:
            request.selected_sample_fields.append(enum_field.value)

        reply_itor = self._datastore_stub.RetrieveSamples(request)
        if not reply_itor:
            raise CogmentError(f"'all_samples' failed to connect")

        try:
            async for reply in reply_itor:
                raw_sample = reply.trial_sample
                param = params[raw_sample.trial_id]
                sample = DatastoreSample(raw_sample, param)
                keep_looping = yield sample
                if keep_looping is not None and not bool(keep_looping):
                    break

        except grpc.aio.AioRpcError as exc:
            logger.debug(f"gRPC failed status details: [{exc.debug_error_string()}]")
            if exc.code() == grpc.StatusCode.UNAVAILABLE:
                logger.error(f"Datastore all_samples communication lost: [{exc.details()}]")
            else:
                logger.exception("Datastore all_samples -- Unexpected aio failure")
                raise

        except GeneratorExit:
            raise

        except asyncio.CancelledError as exc:
            logger.debug(f"Datastore all_samples coroutine cancelled while waiting for samples: [{exc}]")

        except Exception:
            logger.exception("Datastore all_samples")
            raise



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc

from cogment.api import common_pb2 as cogment_dot_api_dot_common__pb2
from cogment.api import model_registry_pb2 as cogment_dot_api_dot_model__registry__pb2


class ModelRegistrySPStub(object):
    """API for model registry, it stores versioned model data and their metadata
    """

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.CreateOrUpdateModel = channel.unary_unary(
                '/cogmentAPI.ModelRegistrySP/CreateOrUpdateModel',
                request_serializer=cogment_dot_api_dot_model__registry__pb2.CreateOrUpdateModelRequest.SerializeToString,
                response_deserializer=cogment_dot_api_dot_model__registry__pb2.CreateOrUpdateModelReply.FromString,
                )
        self.DeleteModel = channel.unary_unary(
                '/cogmentAPI.ModelRegistrySP/DeleteModel',
                request_serializer=cogment_dot_api_dot_model__registry__pb2.DeleteModelRequest.SerializeToString,
                response_deserializer=cogment_dot_api_dot_model__registry__pb2.DeleteModelReply.FromString,
                )
        self.RetrieveModels = channel.unary_unary(
                '/cogmentAPI.ModelRegistrySP/RetrieveModels',
                request_serializer=cogment_dot_api_dot_model__registry__pb2.RetrieveModelsRequest.SerializeToString,
                response_deserializer=cogment_dot_api_dot_model__registry__pb2.RetrieveModelsReply.FromString,
                )
        self.CreateVersion = channel.stream_unary(
                '/cogmentAPI.ModelRegistrySP/CreateVersion',
                request_serializer=cogment_dot_api_dot_model__registry__pb2.CreateVersionRequestChunk.SerializeToString,
                response_deserializer=cogment_dot_api_dot_model__registry__pb2.CreateVersionReply.FromString,
                )
        self.RetrieveVersionInfos = channel.unary_unary(
                '/cogmentAPI.ModelRegistrySP/RetrieveVersionInfos',
                request_serializer=cogment_dot_api_dot_model__registry__pb2.RetrieveVersionInfosRequest.SerializeToString,
                response_deserializer=cogment_dot_api_dot_model__registry__pb2.RetrieveVersionInfosReply.FromString,
                )
        self.RetrieveVersionData = channel.unary_stream(
                '/cogmentAPI.ModelRegistrySP/RetrieveVersionData',
                request_serializer=cogment_dot_api_dot_model__registry__pb2.RetrieveVersionDataRequest.SerializeToString,
                response_deserializer=cogment_dot_api_dot_model__registry__pb2.RetrieveVersionDataReplyChunk.FromString,
                )
        self.Version = channel.unary_unary(
                '/cogmentAPI.ModelRegistrySP/Version',
                request_serializer=cogment_dot_api_dot_common__pb2.VersionRequest.SerializeToString,
                response_deserializer=cogment_dot_api_dot_common__pb2.VersionInfo.FromString,
                )


class ModelRegistrySPServicer(object):
    """API for model registry, it stores versioned model data and their metadata
    """

    def CreateOrUpdateModel(self, request, context):
        """Create or update a model
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def DeleteModel(self, request, context):
        """Delete a model and its versions
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def RetrieveModels(self, request, context):
        """Retrieve the the info of all or part of the models
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def CreateVersion(self, request_iterator, context):
        """Create a model version and returns its information
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def RetrieveVersionInfos(self, request, context):
        """Retrieve the info of all or part of the versions of a model
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def RetrieveVersionData(self, request, context):
        """Retrieve the data of a model given or latest version
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Version(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_ModelRegistrySPServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'CreateOrUpdateModel': grpc.unary_unary_rpc_method_handler(
                    servicer.CreateOrUpdateModel,
                    request_deserializer=cogment_dot_api_dot_model__registry__pb2.CreateOrUpdateModelRequest.FromString,
                    response_serializer=cogment_dot_api_dot_model__registry__pb2.CreateOrUpdateModelReply.SerializeToString,
            ),
            'DeleteModel': grpc.unary_unary_rpc_method_handler(
                    servicer.DeleteModel,
                    request_deserializer=cogment_dot_api_dot_model__registry__pb2.DeleteModelRequest.FromString,
                    response_serializer=cogment_dot_api_dot_model__registry__pb2.DeleteModelReply.SerializeToString,
            ),
            'RetrieveModels': grpc.unary_unary_rpc_method_handler(
                    servicer.RetrieveModels,
                    request_deserializer=cogment_dot_api_dot_model__registry__pb2.RetrieveModelsRequest.FromString,
                    response_serializer=cogment_dot_api_dot_model__registry__pb2.RetrieveModelsReply.SerializeToString,
            ),
            'CreateVersion': grpc.stream_unary_rpc_method_handler(
                    servicer.CreateVersion,
                    request_deserializer=cogment_dot_api_dot_model__registry__pb2.CreateVersionRequestChunk.FromString,
                    response_serializer=cogment_dot_api_dot_model__registry__pb2.CreateVersionReply.SerializeToString,
            ),
            'RetrieveVersionInfos': grpc.unary_unary_rpc_method_handler(
                    servicer.RetrieveVersionInfos,
                    request_deserializer=cogment_dot_api_dot_model__registry__pb2.RetrieveVersionInfosRequest.FromString,
                    response_serializer=cogment_dot_api_dot_model__registry__pb2.RetrieveVersionInfosReply.SerializeToString,
            ),
            'RetrieveVersionData': grpc.unary_stream_rpc_method_handler(
                    servicer.RetrieveVersionData,
                    request_deserializer=cogment_dot_api_dot_model__registry__pb2.RetrieveVersionDataRequest.FromString,
                    response_serializer=cogment_dot_api_dot_model__registry__pb2.RetrieveVersionDataReplyChunk.SerializeToString,
            ),
            'Version': grpc.unary_unary_rpc_method_handler(
                    servicer.Version,
                    request_deserializer=cogment_dot_api_dot_common__pb2.VersionRequest.FromString,
                    response_serializer=cogment_dot_api_dot_common__pb2.VersionInfo.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'cogmentAPI.ModelRegistrySP', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))


 # This class is part of an EXPERIMENTAL API.
class ModelRegistrySP(object):
    """API for model registry, it stores versioned model data and their metadata
    """

    @staticmethod
    def CreateOrUpdateModel(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/cogmentAPI.ModelRegistrySP/CreateOrUpdateModel',
            cogment_dot_api_dot_model__registry__pb2.CreateOrUpdateModelRequest.SerializeToString,
            cogment_dot_api_dot_model__registry__pb2.CreateOrUpdateModelReply.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def DeleteModel(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/cogmentAPI.ModelRegistrySP/DeleteModel',
            cogment_dot_api_dot_model__registry__pb2.DeleteModelRequest.SerializeToString,
            cogment_dot_api_dot_model__registry__pb2.DeleteModelReply.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def RetrieveModels(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/cogmentAPI.ModelRegistrySP/RetrieveModels',
            cogment_dot_api_dot_model__registry__pb2.RetrieveModelsRequest.SerializeToString,
            cogment_dot_api_dot_model__registry__pb2.RetrieveModelsReply.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def CreateVersion(request_iterator,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.stream_unary(request_iterator, target, '/cogmentAPI.ModelRegistrySP/CreateVersion',
            cogment_dot_api_dot_model__registry__pb2.CreateVersionRequestChunk.SerializeToString,
            cogment_dot_api_dot_model__registry__pb2.CreateVersionReply.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def RetrieveVersionInfos(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/cogmentAPI.ModelRegistrySP/RetrieveVersionInfos',
            cogment_dot_api_dot_model__registry__pb2.RetrieveVersionInfosRequest.SerializeToString,
            cogment_dot_api_dot_model__registry__pb2.RetrieveVersionInfosReply.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def RetrieveVersionData(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_stream(request, target, '/cogmentAPI.ModelRegistrySP/RetrieveVersionData',
            cogment_dot_api_dot_model__registry__pb2.RetrieveVersionDataRequest.SerializeToString,
            cogment_dot_api_dot_model__registry__pb2.RetrieveVersionDataReplyChunk.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def Version(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/cogmentAPI.ModelRegistrySP/Version',
            cogment_dot_api_dot_common__pb2.VersionRequest.SerializeToString,
            cogment_dot_api_dot_common__pb2.VersionInfo.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)



# Copyright 2021 AI Redefined Inc. <dev+cogment@ai-r.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from datetime import datetime
from distutils import log
from setuptools.command.build_py import build_py as _build_py
from setuptools.command.develop import develop as _develop
from setuptools.command.sdist import sdist as _sdist
from shutil import unpack_archive, rmtree, copytree
from urllib.request import urlretrieve, urlopen
import json
import os
import pkg_resources
import setuptools

COGMENT_LATEST_VERSION = "latest"


def get_cogment_latest_release_version():
    res = urlopen("https://api.github.com/repos/cogment/cogment/releases/latest")
    parsedBody = json.load(res)
    return parsedBody["tag_name"]


def touch(filename):
    with open(filename, 'w') as fp:
        pass


class RetrieveCogmentApi(setuptools.Command):
    description = "retrieve the cogment api"
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        # This is executed at the root of the package
        package_dir = os.getcwd()

        import yaml

        # In a lot of context the logs are not displayed,
        # putting everything in a string to be able to create a file output
        retrieve_cogment_api_logs = ""

        fh = open(os.path.join(package_dir, ".cogment-api.yaml"), "r")
        cogment_api_cfg = yaml.load(fh, Loader=yaml.FullLoader)

        cogment_version = cogment_api_cfg.get(
            "cogment_version", COGMENT_LATEST_VERSION
        )
        cogment_api_path = cogment_api_cfg.get("cogment_api_path")

        protos_dir = os.path.join(package_dir, "cogment/api")

        if cogment_api_path:
            cogment_api_absolute_path = os.path.normpath(
                os.path.join(package_dir, cogment_api_path)
            )
            # honor the --dry-run flag
            if not self.dry_run:
                rmtree(protos_dir, ignore_errors=True)
                copytree(cogment_api_absolute_path, protos_dir)
                touch(os.path.join(protos_dir, ".gitkeep"))

            log.info(f"copy cogment api from [{cogment_api_absolute_path}]")
        else:
            if cogment_version == COGMENT_LATEST_VERSION:
                cogment_version = get_cogment_latest_release_version()

            url = f"https://github.com/cogment/cogment/releases/download/{cogment_version}/cogment-api.tar.gz"

            # Download the archive to a temp file, will fail if the version doesn't exist
            # or if the network is down thus keeping what was alredy downloaded.
            temp_local_filename, _ = urlretrieve(url)

            # honor the --dry-run flag
            if not self.dry_run:
                rmtree(protos_dir, ignore_errors=True)
                os.makedirs(protos_dir, exist_ok=True)
                touch(os.path.join(protos_dir, ".gitkeep"))

                unpack_archive(temp_local_filename, os.path.join(package_dir, "cogment"), "gztar")

            log.info(f"download cogment api [{cogment_version}] from [{url}]")


class BuildCogmentApiProtos(setuptools.Command):
    description = 'build the cogment api protos'
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        # This is executed at the root of the package
        package_dir = os.getcwd()

        from grpc.tools import protoc

        # Code inspired from
        # https://github.com/grpc/grpc/blob/v1.43.0/tools/distrib/python/grpcio_tools/grpc_tools/command.py#L23-L47
        proto_files = []
        inclusion_root = os.path.abspath(package_dir)
        for root, _, files in os.walk(os.path.join(package_dir, "cogment")):
            for filename in files:
                if filename.endswith('.proto'):
                    proto_files.append(os.path.abspath(os.path.join(root, filename)))

        well_known_protos_include = pkg_resources.resource_filename(
            'grpc_tools', '_proto')

        # honor the --dry-run flag
        if not self.dry_run:
            for proto_file in proto_files:
                command = [
                    'grpc_tools.protoc',
                    '--proto_path={}'.format(inclusion_root),
                    '--proto_path={}'.format(well_known_protos_include),
                    '--python_out={}'.format(inclusion_root),
                    '--grpc_python_out={}'.format(inclusion_root),
                    proto_file
                ]
                if protoc.main(command) != 0:
                    raise Exception('error: {} failed'.format(command))
                log.info(f"generated python implementation of [{proto_file}]")


class develop(_develop):
    def run(self):
        self.run_command('retrieve_cogment_api')
        self.run_command('build_cogment_api_protos')
        super().run()


class sdist(_sdist):
    def run(self):
        self.run_command('retrieve_cogment_api')
        self.run_command('build_cogment_api_protos')
        super().run()


class build_py(_build_py):
    def run(self):
        self.run_command('build_cogment_api_protos')
        super().run()


setuptools.setup(
    setup_requires=[
        "grpcio-tools >=1.42,<1.45",
        "PyYaml ~=5.3.1"
    ],
    cmdclass={
        "retrieve_cogment_api": RetrieveCogmentApi,
        "build_cogment_api_protos": BuildCogmentApiProtos,
        # Executed when the package is installed as a dependency,
        # i.e. when someone does something like pip install cogment.
        "build_py": build_py,
        "develop": develop,
        "sdist": sdist,
    },
)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: cogment/api/health.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x18\x63ogment/api/health.proto\x12\ncogmentAPI\"%\n\x12HealthCheckRequest\x12\x0f\n\x07service\x18\x01 \x01(\t\"\xde\x01\n\x13HealthCheckResponse\x12=\n\x06status\x18\x01 \x01(\x0e\x32-.cogmentAPI.HealthCheckResponse.ServingStatus\x12\x37\n\x0c\x63ogment_data\x18\x64 \x01(\x0b\x32!.cogmentAPI.CogmentHealthMetadata\"O\n\rServingStatus\x12\x0b\n\x07UNKNOWN\x10\x00\x12\x0b\n\x07SERVING\x10\x01\x12\x0f\n\x0bNOT_SERVING\x10\x02\x12\x13\n\x0fSERVICE_UNKNOWN\x10\x03\"\x17\n\x15\x43ogmentHealthMetadata2\x9e\x01\n\x06Health\x12H\n\x05\x43heck\x12\x1e.cogmentAPI.HealthCheckRequest\x1a\x1f.cogmentAPI.HealthCheckResponse\x12J\n\x05Watch\x12\x1e.cogmentAPI.HealthCheckRequest\x1a\x1f.cogmentAPI.HealthCheckResponse0\x01\x62\x06proto3')



_HEALTHCHECKREQUEST = DESCRIPTOR.message_types_by_name['HealthCheckRequest']
_HEALTHCHECKRESPONSE = DESCRIPTOR.message_types_by_name['HealthCheckResponse']
_COGMENTHEALTHMETADATA = DESCRIPTOR.message_types_by_name['CogmentHealthMetadata']
_HEALTHCHECKRESPONSE_SERVINGSTATUS = _HEALTHCHECKRESPONSE.enum_types_by_name['ServingStatus']
HealthCheckRequest = _reflection.GeneratedProtocolMessageType('HealthCheckRequest', (_message.Message,), {
  'DESCRIPTOR' : _HEALTHCHECKREQUEST,
  '__module__' : 'cogment.api.health_pb2'
  # @@protoc_insertion_point(class_scope:cogmentAPI.HealthCheckRequest)
  })
_sym_db.RegisterMessage(HealthCheckRequest)

HealthCheckResponse = _reflection.GeneratedProtocolMessageType('HealthCheckResponse', (_message.Message,), {
  'DESCRIPTOR' : _HEALTHCHECKRESPONSE,
  '__module__' : 'cogment.api.health_pb2'
  # @@protoc_insertion_point(class_scope:cogmentAPI.HealthCheckResponse)
  })
_sym_db.RegisterMessage(HealthCheckResponse)

CogmentHealthMetadata = _reflection.GeneratedProtocolMessageType('CogmentHealthMetadata', (_message.Message,), {
  'DESCRIPTOR' : _COGMENTHEALTHMETADATA,
  '__module__' : 'cogment.api.health_pb2'
  # @@protoc_insertion_point(class_scope:cogmentAPI.CogmentHealthMetadata)
  })
_sym_db.RegisterMessage(CogmentHealthMetadata)

_HEALTH = DESCRIPTOR.services_by_name['Health']
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _HEALTHCHECKREQUEST._serialized_start=40
  _HEALTHCHECKREQUEST._serialized_end=77
  _HEALTHCHECKRESPONSE._serialized_start=80
  _HEALTHCHECKRESPONSE._serialized_end=302
  _HEALTHCHECKRESPONSE_SERVINGSTATUS._serialized_start=223
  _HEALTHCHECKRESPONSE_SERVINGSTATUS._serialized_end=302
  _COGMENTHEALTHMETADATA._serialized_start=304
  _COGMENTHEALTHMETADATA._serialized_end=327
  _HEALTH._serialized_start=330
  _HEALTH._serialized_end=488
# @@protoc_insertion_point(module_scope)



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc

from cogment.api import common_pb2 as cogment_dot_api_dot_common__pb2
from cogment.api import trial_datastore_pb2 as cogment_dot_api_dot_trial__datastore__pb2


class TrialDatastoreSPStub(object):
    """API for trial datastore, it stores trial samples
    -- Online operations (e.g. during training)
    """

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.RetrieveTrials = channel.unary_unary(
                '/cogmentAPI.TrialDatastoreSP/RetrieveTrials',
                request_serializer=cogment_dot_api_dot_trial__datastore__pb2.RetrieveTrialsRequest.SerializeToString,
                response_deserializer=cogment_dot_api_dot_trial__datastore__pb2.RetrieveTrialsReply.FromString,
                )
        self.RetrieveSamples = channel.unary_stream(
                '/cogmentAPI.TrialDatastoreSP/RetrieveSamples',
                request_serializer=cogment_dot_api_dot_trial__datastore__pb2.RetrieveSamplesRequest.SerializeToString,
                response_deserializer=cogment_dot_api_dot_trial__datastore__pb2.RetrieveSampleReply.FromString,
                )
        self.AddTrial = channel.unary_unary(
                '/cogmentAPI.TrialDatastoreSP/AddTrial',
                request_serializer=cogment_dot_api_dot_trial__datastore__pb2.AddTrialRequest.SerializeToString,
                response_deserializer=cogment_dot_api_dot_trial__datastore__pb2.AddTrialReply.FromString,
                )
        self.AddSample = channel.stream_unary(
                '/cogmentAPI.TrialDatastoreSP/AddSample',
                request_serializer=cogment_dot_api_dot_trial__datastore__pb2.AddSampleRequest.SerializeToString,
                response_deserializer=cogment_dot_api_dot_trial__datastore__pb2.AddSamplesReply.FromString,
                )
        self.DeleteTrials = channel.unary_unary(
                '/cogmentAPI.TrialDatastoreSP/DeleteTrials',
                request_serializer=cogment_dot_api_dot_trial__datastore__pb2.DeleteTrialsRequest.SerializeToString,
                response_deserializer=cogment_dot_api_dot_trial__datastore__pb2.DeleteTrialsReply.FromString,
                )
        self.Version = channel.unary_unary(
                '/cogmentAPI.TrialDatastoreSP/Version',
                request_serializer=cogment_dot_api_dot_common__pb2.VersionRequest.SerializeToString,
                response_deserializer=cogment_dot_api_dot_common__pb2.VersionInfo.FromString,
                )


class TrialDatastoreSPServicer(object):
    """API for trial datastore, it stores trial samples
    -- Online operations (e.g. during training)
    """

    def RetrieveTrials(self, request, context):
        """Retrieve the trials matching the given request
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def RetrieveSamples(self, request, context):
        """Retrieve samples from matching trials, trials can be ongoing
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def AddTrial(self, request, context):
        """-- Offline trial data managmement operations

        Add a trial to the activity logger, once a trial is added, samples can be retrieved
        Expected header metadata
        - trial-id
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def AddSample(self, request_iterator, context):
        """Add samples to a trial
        Expected header metadata
        - trial-id
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def DeleteTrials(self, request, context):
        """Delete the trials matching the given request, on failure no trial is deleted
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Version(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_TrialDatastoreSPServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'RetrieveTrials': grpc.unary_unary_rpc_method_handler(
                    servicer.RetrieveTrials,
                    request_deserializer=cogment_dot_api_dot_trial__datastore__pb2.RetrieveTrialsRequest.FromString,
                    response_serializer=cogment_dot_api_dot_trial__datastore__pb2.RetrieveTrialsReply.SerializeToString,
            ),
            'RetrieveSamples': grpc.unary_stream_rpc_method_handler(
                    servicer.RetrieveSamples,
                    request_deserializer=cogment_dot_api_dot_trial__datastore__pb2.RetrieveSamplesRequest.FromString,
                    response_serializer=cogment_dot_api_dot_trial__datastore__pb2.RetrieveSampleReply.SerializeToString,
            ),
            'AddTrial': grpc.unary_unary_rpc_method_handler(
                    servicer.AddTrial,
                    request_deserializer=cogment_dot_api_dot_trial__datastore__pb2.AddTrialRequest.FromString,
                    response_serializer=cogment_dot_api_dot_trial__datastore__pb2.AddTrialReply.SerializeToString,
            ),
            'AddSample': grpc.stream_unary_rpc_method_handler(
                    servicer.AddSample,
                    request_deserializer=cogment_dot_api_dot_trial__datastore__pb2.AddSampleRequest.FromString,
                    response_serializer=cogment_dot_api_dot_trial__datastore__pb2.AddSamplesReply.SerializeToString,
            ),
            'DeleteTrials': grpc.unary_unary_rpc_method_handler(
                    servicer.DeleteTrials,
                    request_deserializer=cogment_dot_api_dot_trial__datastore__pb2.DeleteTrialsRequest.FromString,
                    response_serializer=cogment_dot_api_dot_trial__datastore__pb2.DeleteTrialsReply.SerializeToString,
            ),
            'Version': grpc.unary_unary_rpc_method_handler(
                    servicer.Version,
                    request_deserializer=cogment_dot_api_dot_common__pb2.VersionRequest.FromString,
                    response_serializer=cogment_dot_api_dot_common__pb2.VersionInfo.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'cogmentAPI.TrialDatastoreSP', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))


 # This class is part of an EXPERIMENTAL API.
class TrialDatastoreSP(object):
    """API for trial datastore, it stores trial samples
    -- Online operations (e.g. during training)
    """

    @staticmethod
    def RetrieveTrials(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/cogmentAPI.TrialDatastoreSP/RetrieveTrials',
            cogment_dot_api_dot_trial__datastore__pb2.RetrieveTrialsRequest.SerializeToString,
            cogment_dot_api_dot_trial__datastore__pb2.RetrieveTrialsReply.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def RetrieveSamples(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_stream(request, target, '/cogmentAPI.TrialDatastoreSP/RetrieveSamples',
            cogment_dot_api_dot_trial__datastore__pb2.RetrieveSamplesRequest.SerializeToString,
            cogment_dot_api_dot_trial__datastore__pb2.RetrieveSampleReply.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def AddTrial(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/cogmentAPI.TrialDatastoreSP/AddTrial',
            cogment_dot_api_dot_trial__datastore__pb2.AddTrialRequest.SerializeToString,
            cogment_dot_api_dot_trial__datastore__pb2.AddTrialReply.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def AddSample(request_iterator,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.stream_unary(request_iterator, target, '/cogmentAPI.TrialDatastoreSP/AddSample',
            cogment_dot_api_dot_trial__datastore__pb2.AddSampleRequest.SerializeToString,
            cogment_dot_api_dot_trial__datastore__pb2.AddSamplesReply.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def DeleteTrials(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/cogmentAPI.TrialDatastoreSP/DeleteTrials',
            cogment_dot_api_dot_trial__datastore__pb2.DeleteTrialsRequest.SerializeToString,
            cogment_dot_api_dot_trial__datastore__pb2.DeleteTrialsReply.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def Version(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/cogmentAPI.TrialDatastoreSP/Version',
            cogment_dot_api_dot_common__pb2.VersionRequest.SerializeToString,
            cogment_dot_api_dot_common__pb2.VersionInfo.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)
